from platypush.message.event import Event


class LightStatusChangeEvent(Event):
    """
    Event triggered when the state of a lightbulb changes
    """

    def __init__(self, light_id=None, light_name=None, on=None, bri=None,
                 sat=None, hue=None, ct=None, *args, **kwargs):
        """
        :param light_id: Light ID that triggered the event
        :type light_id: int

        :param light_name: Light name that triggered the event
        :type light_name: str

        :param on: Set if the power state of the bulb changed
        :type on: bool

        :param bri: Set if the brightness state of the bulb changed
        :type bri: int

        :param sat: Set if the saturation state of the bulb changed
        :type sat: int

        :param hue: Set if the hue state of the bulb changed
        :type hue: int

        :param ct: Set if the color temperature state of the bulb changed
        :type ct: int
        """

        attrs = {}

        if light_id is not None:
            attrs['light_id'] = light_id
        if light_name is not None:
            attrs['light_name'] = light_name
        if on is not None:
            attrs['on'] = on
        if bri is not None:
            attrs['bri'] = bri
        if sat is not None:
            attrs['sat'] = sat
        if hue is not None:
            attrs['hue'] = hue
        if ct is not None:
            attrs['ct'] = ct

        super().__init__(*args, **attrs, **kwargs)


# vim:sw=4:ts=4:et:




import random
import statistics
import time

from enum import Enum
from threading import Thread
from redis import Redis
from redis.exceptions import TimeoutError as QueueTimeoutError

from platypush.context import get_backend
from platypush.plugins import action
from platypush.plugins.light import LightPlugin
from platypush.utils import set_thread_name


class LightHuePlugin(LightPlugin):
    """
    Philips Hue lights plugin.

    Requires:

        * **phue** (``pip install phue``)
    """

    MAX_BRI = 255
    MAX_SAT = 255
    MAX_HUE = 65535
    ANIMATION_CTRL_QUEUE_NAME = 'platypush/light/hue/AnimationCtrl'
    _BRIDGE_RECONNECT_SECONDS = 5
    _MAX_RECONNECT_TRIES = 5

    class Animation(Enum):
        COLOR_TRANSITION = 'color_transition'
        BLINK = 'blink'

        def __eq__(self, other):
            if isinstance(other, str):
                return self.value == other
            elif isinstance(other, self.__class__):
                return self == other

    def __init__(self, bridge, lights=None, groups=None):
        """
        :param bridge: Bridge address or hostname
        :type bridge: str

        :param lights: Default lights to be controlled (default: all)
        :type lights: list[str]

        :param groups Default groups to be controlled (default: all)
        :type groups: list[str]
        """

        super().__init__()

        self.bridge_address = bridge
        self.bridge = None
        self.logger.info('Initializing Hue lights plugin - bridge: "{}"'.format(self.bridge_address))

        self.connect()
        self.lights = []; self.groups = []

        if lights:
            self.lights = lights
        elif groups:
            self.groups = groups
            self._expand_groups()
        else:
            self.lights = [l.name for l in self.bridge.lights]

        self.redis = None
        self.animation_thread = None
        self.animations = {}
        self._init_animations()
        self.logger.info('Configured lights: "{}"'. format(self.lights))

    def _expand_groups(self):
        groups = [g for g in self.bridge.groups if g.name in self.groups]
        for g in groups:
            for l in g.lights:
                self.lights += [l.name]

    def _init_animations(self):
        self.animations = {
            'groups': {},
            'lights': {},
        }

        for g in self.bridge.groups:
            self.animations['groups'][g.group_id] = None
        for l in self.bridge.lights:
            self.animations['lights'][l.light_id] = None

    @action
    def connect(self):
        """
        Connect to the configured Hue bridge. If the device hasn't been paired
        yet, uncomment the ``.connect()`` and ``.get_api()`` lines and retry
        after clicking the pairing button on your bridge.
        """

        # Lazy init
        if not self.bridge:
            from phue import Bridge, PhueRegistrationException
            success = False
            n_tries = 0

            while not success:
                try:
                    n_tries += 1
                    self.bridge = Bridge(self.bridge_address)
                    success = True
                except PhueRegistrationException as e:
                    self.logger.warning('Bridge registration error: {}'.
                                        format(str(e)))

                    if n_tries >= self._MAX_RECONNECT_TRIES:
                        self.logger.error(('Bridge registration failed after ' +
                                           '{} attempts').format(n_tries))
                        break

                    time.sleep(self._BRIDGE_RECONNECT_SECONDS)

            self.logger.info('Bridge connected')
            self.get_scenes()
        else:
            self.logger.info('Bridge already connected')

    @action
    def get_scenes(self):
        """
        Get the available scenes on the devices.

        :returns: The scenes configured on the bridge.

        Example output::

            {
                "scene-id-1": {
                    "name": "Scene 1",
                    "lights": [
                        "1",
                        "3"
                    ],

                    "owner": "owner-id",
                    "recycle": true,
                    "locked": false,
                    "appdata": {},
                    "picture": "",
                    "lastupdated": "2018-06-01T00:00:00",
                    "version": 1
                }
            }

        """

        return self.bridge.get_scene()

    @action
    def get_lights(self):
        """
        Get the configured lights.

        :returns: List of available lights as id->dict.

        Example::

            {
                "1": {
                    "state": {
                        "on": true,
                        "bri": 254,
                        "hue": 1532,
                        "sat": 215,
                        "effect": "none",
                        "xy": [
                            0.6163,
                            0.3403
                        ],

                        "ct": 153,
                        "alert": "none",
                        "colormode": "hs",
                        "reachable": true
                    },

                    "type": "Extended color light",
                    "name": "Lightbulb 1",
                    "modelid": "LCT001",
                    "manufacturername": "Philips",
                    "uniqueid": "00:11:22:33:44:55:66:77-88",
                    "swversion": "5.105.0.21169"
                }
            }

        """

        return self.bridge.get_light()

    @action
    def get_groups(self):
        """
        Get the list of configured light groups.

        :returns: List of configured light groups as id->dict.

        Example::

            {
                "1": {
                    "name": "Living Room",
                    "lights": [
                        "16",
                        "13",
                        "12",
                        "11",
                        "10",
                        "9",
                        "1",
                        "3"
                    ],

                    "type": "Room",
                    "state": {
                        "all_on": true,
                        "any_on": true
                    },

                    "class": "Living room",
                    "action": {
                        "on": true,
                        "bri": 241,
                        "hue": 37947,
                        "sat": 221,
                        "effect": "none",
                        "xy": [
                            0.2844,
                            0.2609
                        ],

                        "ct": 153,
                        "alert": "none",
                        "colormode": "hs"
                    }
                }
            }

        """

        return self.bridge.get_group()

    @action
    def get_animations(self):
        """
        Get the list of running light animations.

        :returns: dict.

        Structure::

            {
                "groups": {
                    "id_1": {
                        "type": "color_transition",
                        "hue_range": [0,65535],
                        "sat_range": [0,255],
                        "bri_range": [0,255],
                        "hue_step": 10,
                        "sat_step": 10,
                        "bri_step": 2,
                        "transition_seconds": 2

                    }

                },

                "lights": {
                    "id_1": {}

                }

            }

        """

        return self.animations

    def _exec(self, attr, *args, **kwargs):
        try:
            self.connect()
            self.stop_animation()
        except Exception as e:
            # Reset bridge connection
            self.bridge = None
            raise e

        lights = []; groups = []
        if 'lights' in kwargs:
            lights = kwargs.pop('lights').split(',').strip() \
                if isinstance(lights, str) else kwargs.pop('lights')
        if 'groups' in kwargs:
            groups = kwargs.pop('groups').split(',').strip() \
                if isinstance(groups, str) else kwargs.pop('groups')

        if not lights and not groups:
            lights = self.lights
            groups = self.groups

        if not self.bridge:
            self.connect()

        try:
            if attr == 'scene':
                self.bridge.run_scene(groups[0], kwargs.pop('name'))
            else:
                if groups:
                    self.bridge.set_group(groups, attr, *args, **kwargs)
                if lights:
                    self.bridge.set_light(lights, attr, *args, **kwargs)
        except Exception as e:
            # Reset bridge connection
            self.bridge = None
            raise e

    @action
    def set_light(self, light, **kwargs):
        """
        Set a light (or lights) property.

        :param light: Light or lights to set. Can be a string representing the light name,
            a light object, a list of string, or a list of light objects.
        :param kwargs: key-value list of parameters to set.

        Example call::

            {
                "type": "request",
                "target": "hostname",
                "action": "light.hue.set_light",
                "args": {
                    "light": "Bulb 1",
                    "sat": 255
                }
            }

        """

        self.connect()
        self.bridge.set_light(light, **kwargs)

    @action
    def set_group(self, group, **kwargs):
        """
        Set a group (or groups) property.

        :param group: Group or groups to set. Can be a string representing the group name, a group object, a list of strings, or a list of group objects.
        :param kwargs: key-value list of parameters to set.

        Example call::

            {
                "type": "request",
                "target": "hostname",
                "action": "light.hue.set_group",
                "args": {
                    "light": "Living Room",
                    "sat": 255
                }
            }

        """

        self.connect()
        self.bridge.set_group(group, **kwargs)

    @action
    def on(self, lights=None, groups=None, **kwargs):
        """
        Turn lights/groups on.

        :param lights: Lights to turn on (names or light objects). Default: plugin default lights
        :param groups: Groups to turn on (names or group objects). Default: plugin default groups
        """

        if groups is None:
            groups = []
        if lights is None:
            lights = []
        return self._exec('on', True, lights=lights, groups=groups, **kwargs)

    @action
    def off(self, lights=None, groups=None, **kwargs):
        """
        Turn lights/groups off.

        :param lights: Lights to turn off (names or light objects). Default: plugin default lights
        :param groups: Groups to turn off (names or group objects). Default: plugin default groups
        """

        if groups is None:
            groups = []
        if lights is None:
            lights = []
        return self._exec('on', False, lights=lights, groups=groups, **kwargs)

    @action
    def toggle(self, lights=None, groups=None, **kwargs):
        """
        Toggle lights/groups on/off.

        :param lights: Lights to turn off (names or light objects). Default: plugin default lights
        :param groups: Groups to turn off (names or group objects). Default: plugin default groups
        """

        if groups is None:
            groups = []
        if lights is None:
            lights = []
        lights_on  = []
        lights_off = []
        groups_on  = []
        groups_off = []

        if groups:
            all_groups = self.bridge.get_group().values()

            groups_on = [
                group['name'] for group in all_groups
                if group['name'] in groups and group['state']['any_on'] is True
            ]

            groups_off = [
                group['name'] for group in all_groups
                if group['name'] in groups and group['state']['any_on'] is False
            ]

        if not groups and not lights:
            lights = self.lights

        if lights:
            all_lights = self.bridge.get_light().values()

            lights_on = [
                light['name'] for light in all_lights
                if light['name'] in lights and light['state']['on'] is True
            ]

            lights_off = [
                light['name'] for light in all_lights
                if light['name'] in lights and light['state']['on'] is False
            ]

        if lights_on or groups_on:
            self._exec('on', False, lights=lights_on, groups=groups_on, **kwargs)

        if lights_off or groups_off:
            self._exec('on', True, lights=lights_off, groups=groups_off, **kwargs)

    @action
    def bri(self, value, lights=None, groups=None, **kwargs):
        """
        Set lights/groups brightness.

        :param lights: Lights to control (names or light objects). Default: plugin default lights
        :param groups: Groups to control (names or group objects). Default: plugin default groups
        :param value: Brightness value (range: 0-255)
        """

        if groups is None:
            groups = []
        if lights is None:
            lights = []
        return self._exec('bri', int(value) % (self.MAX_BRI+1),
                          lights=lights, groups=groups, **kwargs)

    @action
    def sat(self, value, lights=None, groups=None, **kwargs):
        """
        Set lights/groups saturation.

        :param lights: Lights to control (names or light objects). Default: plugin default lights
        :param groups: Groups to control (names or group objects). Default: plugin default groups
        :param value: Saturation value (range: 0-255)
        """

        if groups is None:
            groups = []
        if lights is None:
            lights = []
        return self._exec('sat', int(value) % (self.MAX_SAT+1),
                          lights=lights, groups=groups, **kwargs)

    @action
    def hue(self, value, lights=None, groups=None, **kwargs):
        """
        Set lights/groups color hue.

        :param lights: Lights to control (names or light objects). Default: plugin default lights
        :param groups: Groups to control (names or group objects). Default: plugin default groups
        :param value: Hue value (range: 0-65535)
        """

        if groups is None:
            groups = []
        if lights is None:
            lights = []
        return self._exec('hue', int(value) % (self.MAX_HUE+1),
                          lights=lights, groups=groups, **kwargs)

    @action
    def xy(self, value, lights=None, groups=None, **kwargs):
        """
        Set lights/groups XY colors.

        :param value: xY value
        :type value: list[float] containing the two values
        """
        if groups is None:
            groups = []
        if lights is None:
            lights = []
        return self._exec('xy', value, lights=lights, groups=groups, **kwargs)

    @action
    def ct(self, value, lights=None, groups=None, **kwargs):
        """
        Set lights/groups color temperature.

        :param value: Temperature value (range: 0-255)
        :type value: int
        """
        if groups is None:
            groups = []
        if lights is None:
            lights = []
        return self._exec('ct', value, lights=lights, groups=groups, **kwargs)

    @action
    def delta_bri(self, delta, lights=None, groups=None, **kwargs):
        """
        Change lights/groups brightness by a delta [-100, 100] compared to the current state.

        :param lights: Lights to control (names or light objects). Default: plugin default lights
        :param groups: Groups to control (names or group objects). Default: plugin default groups
        :param delta: Brightness delta value (range: -100, 100)
        """

        if groups is None:
            groups = []
        if lights is None:
            lights = []
        bri = 0

        if lights:
            bri = statistics.mean([
                light['state']['bri']
                for light in self.bridge.get_light().values()
                if light['name'] in lights
            ])
        elif groups:
            bri = statistics.mean([
                group['action']['bri']
                for group in self.bridge.get_group().values()
                if group['name'] in groups
            ])
        else:
            bri = statistics.mean([
                light['state']['bri']
                for light in self.bridge.get_light().values()
                if light['name'] in self.lights
            ])

        delta *= (self.MAX_BRI/100)
        if bri+delta < 0:
            bri = 0
        elif bri+delta > self.MAX_BRI:
            bri = self.MAX_BRI
        else:
            bri += delta

        return self._exec('bri', int(bri), lights=lights, groups=groups, **kwargs)

    @action
    def delta_sat(self, delta, lights=None, groups=None, **kwargs):
        """
        Change lights/groups saturation by a delta [-100, 100] compared to the current state.

        :param lights: Lights to control (names or light objects). Default: plugin default lights
        :param groups: Groups to control (names or group objects). Default: plugin default groups
        :param delta: Saturation delta value (range: -100, 100)
        """

        if groups is None:
            groups = []
        if lights is None:
            lights = []
        sat = 0

        if lights:
            sat = statistics.mean([
                light['state']['sat']
                for light in self.bridge.get_light().values()
                if light['name'] in lights
            ])
        elif groups:
            sat = statistics.mean([
                group['action']['sat']
                for group in self.bridge.get_group().values()
                if group['name'] in groups
            ])
        else:
            sat = statistics.mean([
                light['state']['sat']
                for light in self.bridge.get_light().values()
                if light['name'] in self.lights
            ])

        delta *= (self.MAX_SAT/100)
        if sat+delta < 0:
            sat = 0
        elif sat+delta > self.MAX_SAT:
            sat = self.MAX_SAT
        else:
            sat += delta

        return self._exec('sat', int(sat), lights=lights, groups=groups, **kwargs)

    @action
    def delta_hue(self, delta, lights=None, groups=None, **kwargs):
        """
        Change lights/groups hue by a delta [-100, 100] compared to the current state.

        :param lights: Lights to control (names or light objects). Default: plugin default lights
        :param groups: Groups to control (names or group objects). Default: plugin default groups
        :param delta: Hue delta value (range: -100, 100)
        """

        if groups is None:
            groups = []
        if lights is None:
            lights = []
        hue = 0

        if lights:
            hue = statistics.mean([
                light['state']['hue']
                for light in self.bridge.get_light().values()
                if light['name'] in lights
            ])
        elif groups:
            hue = statistics.mean([
                group['action']['hue']
                for group in self.bridge.get_group().values()
                if group['name'] in groups
            ])
        else:
            hue = statistics.mean([
                light['state']['hue']
                for light in self.bridge.get_light().values()
                if light['name'] in self.lights
            ])

        delta *= (self.MAX_HUE/100)
        if hue+delta < 0:
            hue = 0
        elif hue+delta > self.MAX_HUE:
            hue = self.MAX_HUE
        else:
            hue += delta

        return self._exec('hue', int(hue), lights=lights, groups=groups, **kwargs)

    @action
    def scene(self, name, lights=None, groups=None, **kwargs):
        """
        Set a scene by name.

        :param lights: Lights to control (names or light objects). Default: plugin default lights
        :param groups: Groups to control (names or group objects). Default: plugin default groups
        :param name: Name of the scene
        """

        if groups is None:
            groups = []
        if lights is None:
            lights = []
        return self._exec('scene', name=name, lights=lights, groups=groups, **kwargs)

    @action
    def is_animation_running(self):
        """
        :returns: True if there is an animation running, false otherwise.
        """

        return self.animation_thread is not None

    @action
    def stop_animation(self):
        """
        Stop a running animation if any
        """

        if self.animation_thread and self.animation_thread.is_alive():
            redis = self._get_redis()
            redis.rpush(self.ANIMATION_CTRL_QUEUE_NAME, 'STOP')
            self._init_animations()

    @action
    def animate(self, animation, duration=None,
                hue_range=None, sat_range=None,
                bri_range=None, lights=None, groups=None,
                hue_step=1000, sat_step=2, bri_step=1, transition_seconds=1.0):
        """
        Run a lights animation.

        :param animation: Animation name. Supported types: **color_transition** and **blink**
        :type animation: str

        :param duration: Animation duration in seconds (default: None, i.e. continue until stop)
        :type duration: float

        :param hue_range: If you selected a color color_transition.html, this will specify the hue range of your color color_transition.html.
            Default: [0, 65535]
        :type hue_range: list[int]

        :param sat_range: If you selected a color color_transition.html, this will specify the saturation range of your color
            color_transition.html. Default: [0, 255]
        :type sat_range: list[int]

        :param bri_range: If you selected a color color_transition.html, this will specify the brightness range of your color
            color_transition.html. Default: [254, 255] :type bri_range: list[int]

        :param lights: Lights to control (names, IDs or light objects). Default: plugin default lights
        :param groups: Groups to control (names, IDs or group objects). Default: plugin default groups

        :param hue_step: If you selected a color color_transition.html, this will specify by how much the color hue will change
            between iterations. Default: 1000 :type hue_step: int

        :param sat_step: If you selected a color color_transition.html, this will specify by how much the saturation will change
            between iterations. Default: 2 :type sat_step: int

        :param bri_step: If you selected a color color_transition.html, this will specify by how much the brightness will change
            between iterations. Default: 1 :type bri_step: int

        :param transition_seconds: Time between two transitions or blinks in seconds. Default: 1.0
        :type transition_seconds: float
        """

        if bri_range is None:
            bri_range = [self.MAX_BRI - 1, self.MAX_BRI]
        if sat_range is None:
            sat_range = [0, self.MAX_SAT]
        if hue_range is None:
            hue_range = [0, self.MAX_HUE]
        if groups:
            groups = [g for g in self.bridge.groups if g.name in groups or g.group_id in groups]
            lights = lights or []
            for g in groups:
                lights.extend([l.name for l in g.lights])
        elif lights:
            lights = [l.name for l in self.bridge.lights if l.name in lights or l.light_id in lights]
        else:
            lights = self.lights

        info = {
            'type': animation,
            'duration': duration,
            'hue_range': hue_range,
            'sat_range': sat_range,
            'bri_range': bri_range,
            'hue_step': hue_step,
            'sat_step': sat_step,
            'bri_step': bri_step,
            'transition_seconds': transition_seconds,
        }

        if groups:
            for g in groups:
                self.animations['groups'][g.group_id] = info

        for l in self.bridge.lights:
            if l.name in lights:
                self.animations['lights'][l.light_id] = info

        def _initialize_light_attrs(lights):
            if animation == self.Animation.COLOR_TRANSITION:
                return { l: {
                    'hue': random.randint(hue_range[0], hue_range[1]),
                    'sat': random.randint(sat_range[0], sat_range[1]),
                    'bri': random.randint(bri_range[0], bri_range[1]),
                } for l in lights }
            elif animation == self.Animation.BLINK:
                return { l: {
                    'on': True,
                    'bri': self.MAX_BRI,
                    'transitiontime': 0,
                } for l in lights }

        def _next_light_attrs(lights):
            if animation == self.Animation.COLOR_TRANSITION:
                for (light, attrs) in lights.items():
                    for (attr, value) in attrs.items():
                        attr_range = [0,0]
                        attr_step = 0

                        if attr == 'hue':
                            attr_range = hue_range
                            attr_step = hue_step
                        elif attr == 'bri':
                            attr_range = bri_range
                            attr_step = bri_step
                        elif attr == 'sat':
                            attr_range = sat_range
                            attr_step = sat_step

                        lights[light][attr] = ((value - attr_range[0] + attr_step) %
                                                (attr_range[1]-attr_range[0]+1)) + \
                                                attr_range[0]
            elif animation == self.Animation.BLINK:
                lights = { light: {
                    'on': False if attrs['on'] else True,
                    'bri': self.MAX_BRI,
                    'transitiontime': 0,
                } for (light, attrs) in lights.items() }

            return lights

        def _should_stop():
            try:
                redis = self._get_redis(transition_seconds)
                redis.blpop(self.ANIMATION_CTRL_QUEUE_NAME)
                return True
            except QueueTimeoutError:
                return False

        def _animate_thread(lights):
            set_thread_name('HueAnimate')
            self.logger.info('Starting {} animation'.format(
                animation, (lights or groups)))

            lights = _initialize_light_attrs(lights)
            animation_start_time = time.time()
            stop_animation = False

            while True:
                if stop_animation or \
                        (duration and time.time() - animation_start_time > duration):
                    break

                try:
                    if animation == self.Animation.COLOR_TRANSITION:
                        for (light, attrs) in lights.items():
                            self.logger.debug('Setting {} to {}'.format(light, attrs))
                            self.bridge.set_light(light, attrs)
                            stop_animation = _should_stop()
                            if stop_animation: break
                    elif animation == self.Animation.BLINK:
                        conf = lights[list(lights.keys())[0]]
                        self.logger.debug('Setting lights to {}'.format(conf))

                        if groups:
                            self.bridge.set_group([g.name for g in groups], conf)
                        else:
                            self.bridge.set_light(lights.keys(), conf)

                        stop_animation = _should_stop()
                        if stop_animation: break
                except Exception as e:
                    self.logger.warning(e)
                    time.sleep(2)

                lights = _next_light_attrs(lights)

            self.logger.info('Stopping animation')
            self.animation_thread = None
            self.redis = None

        self.stop_animation()
        self.animation_thread = Thread(target=_animate_thread,
                                       name='HueAnimate',
                                       args=(lights,))
        self.animation_thread.start()

    def _get_redis(self, socket_timeout=1.0):
        if not self.redis:
            redis_args = get_backend('redis').redis_args
            redis_args['socket_timeout'] = socket_timeout
            self.redis = Redis(**redis_args)
        return self.redis

    def status(self):
        # TODO
        pass


# vim:sw=4:ts=4:et:



"""
.. moduleauthor:: Fabio Manganiello <blacklight86@gmail.com>
"""

import os
import queue
import stat
import tempfile
import time

from enum import Enum
from threading import Thread, Event, RLock

from .core import Sound, Mix

from platypush.context import get_bus
from platypush.message.event.sound import \
    SoundRecordingStartedEvent, SoundRecordingStoppedEvent

from platypush.plugins import Plugin, action


class PlaybackState(Enum):
    STOPPED = 'STOPPED',
    PLAYING = 'PLAYING',
    PAUSED = 'PAUSED'


class RecordingState(Enum):
    STOPPED = 'STOPPED',
    RECORDING = 'RECORDING',
    PAUSED = 'PAUSED'


class SoundPlugin(Plugin):
    """
    Plugin to interact with a sound device.

    Triggers:

        * :class:`platypush.message.event.sound.SoundPlaybackStartedEvent` on playback start
        * :class:`platypush.message.event.sound.SoundPlaybackStoppedEvent` on playback stop
        * :class:`platypush.message.event.sound.SoundPlaybackPausedEvent` on playback pause
        * :class:`platypush.message.event.sound.SoundRecordingStartedEvent` on recording start
        * :class:`platypush.message.event.sound.SoundRecordingStoppedEvent` on recording stop
        * :class:`platypush.message.event.sound.SoundRecordingPausedEvent` on recording pause

    Requires:

        * **sounddevice** (``pip install sounddevice``)
        * **soundfile** (``pip install soundfile``)
        * **numpy** (``pip install numpy``)
    """

    _STREAM_NAME_PREFIX = 'platypush-stream-'
    _default_input_stream_fifo = os.path.join(tempfile.gettempdir(), 'inputstream')

    # noinspection PyProtectedMember
    def __init__(self, input_device=None, output_device=None,
                 input_blocksize=Sound._DEFAULT_BLOCKSIZE,
                 output_blocksize=Sound._DEFAULT_BLOCKSIZE, **kwargs):
        """
        :param input_device: Index or name of the default input device. Use
            :meth:`platypush.plugins.sound.query_devices` to get the
            available devices. Default: system default
        :type input_device: int or str

        :param output_device: Index or name of the default output device.
            Use :meth:`platypush.plugins.sound.query_devices` to get the
            available devices. Default: system default
        :type output_device: int or str

        :param input_blocksize: Blocksize to be applied to the input device.
            Try to increase this value if you get input overflow errors while
            recording. Default: 1024
        :type input_blocksize: int

        :param output_blocksize: Blocksize to be applied to the output device.
            Try to increase this value if you get output underflow errors while
            playing. Default: 1024
        :type output_blocksize: int
        """

        super().__init__(**kwargs)

        self.input_device = input_device
        self.output_device = output_device
        self.input_blocksize = input_blocksize
        self.output_blocksize = output_blocksize

        self.playback_state = {}
        self.playback_state_lock = RLock()
        self.playback_paused_changed = {}
        self.stream_mixes = {}
        self.recording_state = RecordingState.STOPPED
        self.recording_state_lock = RLock()
        self.recording_paused_changed = Event()
        self.active_streams = {}
        self.stream_name_to_index = {}
        self.stream_index_to_name = {}
        self.completed_callback_events = {}

    @staticmethod
    def _get_default_device(category):
        """
        Query the default audio devices.

        :param category: Device category to query. Can be either input or output
        :type category: str
        """

        import sounddevice as sd
        return sd.query_hostapis()[0].get('default_' + category.lower() + '_device')

    @action
    def query_devices(self, category=None):
        """
        Query the available devices

        :param category: Device category to query. Can be either input or output. Default: None (query all devices)
        :type category: str

        :returns: A dictionary representing the available devices.

        Example::

            [
                {
                    "name": "pulse",
                    "hostapi": 0,
                    "max_input_channels": 32,
                    "max_output_channels": 32,
                    "default_low_input_latency": 0.008684807256235827,
                    "default_low_output_latency": 0.008684807256235827,
                    "default_high_input_latency": 0.034807256235827665,
                    "default_high_output_latency": 0.034807256235827665,
                    "default_samplerate": 44100
                },
                {
                    "name": "default",
                    "hostapi": 0,
                    "max_input_channels": 32,
                    "max_output_channels": 32,
                    "default_low_input_latency": 0.008684807256235827,
                    "default_low_output_latency": 0.008684807256235827,
                    "default_high_input_latency": 0.034807256235827665,
                    "default_high_output_latency": 0.034807256235827665,
                    "default_samplerate": 44100
                }
            ]

        """

        import sounddevice as sd

        devs = sd.query_devices()
        if category == 'input':
            devs = [d for d in devs if d.get('max_input_channels') > 0]
        elif category == 'output':
            devs = [d for d in devs if d.get('max_output_channels') > 0]

        return devs

    def _play_audio_callback(self, q, blocksize, streamtype, stream_index):
        import sounddevice as sd

        is_raw_stream = streamtype == sd.RawOutputStream

        # noinspection PyUnusedLocal
        def audio_callback(outdata, frames, frame_time, status):
            if self._get_playback_state(stream_index) == PlaybackState.STOPPED:
                raise sd.CallbackStop

            while self._get_playback_state(stream_index) == PlaybackState.PAUSED:
                self.playback_paused_changed[stream_index].wait()

            if frames != blocksize:
                self.logger.warning('Received {} frames, expected blocksize is {}'.
                                    format(frames, blocksize))
                return

            if status.output_underflow:
                self.logger.warning('Output underflow: increase blocksize?')
                outdata = (b'\x00' if is_raw_stream else 0.) * len(outdata)
                return

            if status:
                self.logger.warning('Audio callback failed: {}'.format(status))

            try:
                data = q.get_nowait()
            except queue.Empty:
                self.logger.warning('Buffer is empty: increase buffersize?')
                raise sd.CallbackStop

            if len(data) < len(outdata):
                outdata[:len(data)] = data
                outdata[len(data):] = (b'\x00' if is_raw_stream else 0.) * \
                                      (len(outdata) - len(data))
            else:
                outdata[:] = data

        return audio_callback

    @action
    def play(self, file=None, sound=None, device=None, blocksize=None,
             bufsize=None, samplerate=None, channels=None, stream_name=None,
             stream_index=None):
        """
        Plays a sound file (support formats: wav, raw) or a synthetic sound.

        :param file: Sound file path. Specify this if you want to play a file
        :type file: str

        :param sound: Sound to play. Specify this if you want to play
            synthetic sounds. You can also create polyphonic sounds by just
            calling play multiple times.
        :type sound: Sound. You can initialize it either from a list
            of `Sound` objects or from its JSON representation, e.g.::

                {
                    "midi_note": 69,  # 440 Hz A
                    "gain":      1.0, # Maximum volume
                    "duration":  1.0  # 1 second or until release/pause/stop
                }

        :param device: Output device (default: default configured device or
            system default audio output if not configured)
        :type device: int or str

        :param blocksize: Audio block size (default: configured
            `output_blocksize` or 2048)
        :type blocksize: int

        :param bufsize: Size of the audio buffer (default: 20 frames for audio
            files, 2 frames for synth sounds)
        :type bufsize: int

        :param samplerate: Audio samplerate. Default: audio file samplerate if
            in file mode, 44100 Hz if in synth mode
        :type samplerate: int

        :param channels: Number of audio channels. Default: number of channels
            in the audio file in file mode, 1 if in synth mode
        :type channels: int

        :param stream_index: If specified, play to an already active stream
            index (you can get them through
            :meth:`platypush.plugins.sound.query_streams`). Default:
            creates a new audio stream through PortAudio.
        :type stream_index: int

        :param stream_name: Name of the stream to play to. If set, the sound
            will be played to the specified stream name, or a stream with that
            name will be created. If not set, and ``stream_index`` is not set
            either, then a new stream will be created on the next available
            index and named ``platypush-stream-<index>``.
        :type stream_name: str
        """

        if not file and not sound:
            raise RuntimeError('Please specify either a file to play or a ' +
                               'list of sound objects')

        import sounddevice as sd

        if blocksize is None:
            blocksize = self.output_blocksize

        if bufsize is None:
            if file:
                bufsize = Sound._DEFAULT_FILE_BUFSIZE
            else:
                bufsize = Sound._DEFAULT_SYNTH_BUFSIZE

        q = queue.Queue(maxsize=bufsize)
        f = None
        t = 0.

        if file:
            file = os.path.abspath(os.path.expanduser(file))

        if device is None:
            device = self.output_device
        if device is None:
            device = self._get_default_device('output')

        if file:
            import soundfile as sf
            f = sf.SoundFile(file)
        if not samplerate:
            samplerate = f.samplerate if f else Sound._DEFAULT_SAMPLERATE
        if not channels:
            channels = f.channels if f else 1

        mix = None
        with self.playback_state_lock:
            stream_index, is_new_stream = self._get_or_allocate_stream_index(
                stream_index=stream_index, stream_name=stream_name)

            if sound and stream_index in self.stream_mixes:
                mix = self.stream_mixes[stream_index]
                mix.add(sound)

        if not mix:
            return None, "Unable to allocate the stream"

        self.logger.info(('Starting playback of {} to sound device [{}] ' +
                          'on stream [{}]').format(
            file or sound, device, stream_index))

        if not is_new_stream:
            return  # Let the existing callback handle the new mix
            # TODO Potential support also for mixed streams with
            # multiple sound files and synth sounds?

        try:
            # Audio queue pre-fill loop
            for _ in range(bufsize):
                if f:
                    data = f.buffer_read(blocksize, dtype='float32')
                    if not data:
                        break
                else:
                    duration = mix.duration()
                    blocktime = float(blocksize / samplerate)
                    next_t = min(t + blocktime, duration) \
                        if duration is not None else t + blocktime

                    data = mix.get_wave(t_start=t, t_end=next_t, samplerate=samplerate)
                    t = next_t

                    if duration is not None and t >= duration:
                        break

                q.put_nowait(data)  # Pre-fill the audio queue

            stream = self.active_streams[stream_index]
            completed_callback_event = self.completed_callback_events[stream_index]

            if stream is None:
                streamtype = sd.RawOutputStream if file else sd.OutputStream
                stream = streamtype(samplerate=samplerate, blocksize=blocksize,
                                    device=device, channels=channels,
                                    dtype='float32',
                                    callback=self._play_audio_callback(
                                        q=q, blocksize=blocksize,
                                        streamtype=streamtype,
                                        stream_index=stream_index),
                                    finished_callback=completed_callback_event.set)

                self._start_playback(stream_index=stream_index, stream=stream)

            with stream:
                # Timeout set until we expect all the buffered blocks to
                # be consumed
                timeout = blocksize * bufsize / samplerate

                while True:
                    while self._get_playback_state(stream_index) == \
                            PlaybackState.PAUSED:
                        self.playback_paused_changed[stream_index].wait()

                    if f:
                        data = f.buffer_read(blocksize, dtype='float32')
                        if not data:
                            break
                    else:
                        duration = mix.duration()
                        blocktime = float(blocksize / samplerate)
                        next_t = min(t + blocktime, duration) \
                            if duration is not None else t + blocktime

                        data = mix.get_wave(t_start=t, t_end=next_t,
                                            samplerate=samplerate)
                        t = next_t

                        if duration is not None and t >= duration:
                            break

                    if self._get_playback_state(stream_index) == \
                            PlaybackState.STOPPED:
                        break

                    try:
                        q.put(data, timeout=timeout)
                    except queue.Full as e:
                        if self._get_playback_state(stream_index) != \
                                PlaybackState.PAUSED:
                            raise e

                completed_callback_event.wait()
        except queue.Full as e:
            if stream_index is None or \
                    self._get_playback_state(stream_index) != PlaybackState.STOPPED:
                self.logger.warning('Playback timeout: audio callback failed?')
        finally:
            if f and not f.closed:
                f.close()
                f = None

            self.stop_playback([stream_index])

    @action
    def stream_recording(self, device=None, fifo=None, duration=None, sample_rate=None,
                         dtype='float32', blocksize=None, latency=0, channels=1):
        """
        Return audio data from an audio source

        :param device: Input device (default: default configured device or system default audio input if not configured)
        :type device: int or str

        :param fifo: Path of the FIFO that will be used to exchange audio samples (default: /tmp/inputstream)
        :type fifo: str

        :param duration: Recording duration in seconds (default: record until stop event)
        :type duration: float

        :param sample_rate: Recording sample rate (default: device default rate)
        :type sample_rate: int

        :param dtype: Data type for the audio samples. Supported types:
            'float64', 'float32', 'int32', 'int16', 'int8', 'uint8'. Default: float32
        :type dtype: str

        :param blocksize: Audio block size (default: configured `input_blocksize` or 2048)
        :type blocksize: int

        :param latency: Device latency in seconds (default: 0)
        :type latency: float

        :param channels: Number of channels (default: 1)
        :type channels: int
        """

        import sounddevice as sd

        self.recording_paused_changed.clear()

        if device is None:
            device = self.input_device
        if device is None:
            device = self._get_default_device('input')

        if sample_rate is None:
            dev_info = sd.query_devices(device, 'input')
            sample_rate = int(dev_info['default_samplerate'])

        if blocksize is None:
            blocksize = self.input_blocksize

        if not fifo:
            fifo = self._default_input_stream_fifo

        q = queue.Queue()

        # noinspection PyUnusedLocal
        def audio_callback(indata, frames, time_duration, status):
            while self._get_recording_state() == RecordingState.PAUSED:
                self.recording_paused_changed.wait()

            if status:
                self.logger.warning('Recording callback status: {}'.format(str(status)))

            q.put(indata.copy())

        def streaming_thread():
            try:
                with sd.InputStream(samplerate=sample_rate, device=device,
                                    channels=channels, callback=audio_callback,
                                    dtype=dtype, latency=latency, blocksize=blocksize):
                    with open(fifo, 'wb') as audio_queue:
                        self.start_recording()
                        get_bus().post(SoundRecordingStartedEvent())
                        self.logger.info('Started recording from device [{}]'.format(device))
                        recording_started_time = time.time()

                        while self._get_recording_state() != RecordingState.STOPPED \
                                and (duration is None or
                                     time.time() - recording_started_time < duration):
                            while self._get_recording_state() == RecordingState.PAUSED:
                                self.recording_paused_changed.wait()

                            get_args = {
                                'block': True,
                                'timeout': max(0, duration - (time.time() - recording_started_time)),
                            } if duration is not None else {}

                            data = q.get(**get_args)
                            if not len(data):
                                continue

                            audio_queue.write(data)
            except queue.Empty:
                self.logger.warning('Recording timeout: audio callback failed?')
            finally:
                self.stop_recording()
                get_bus().post(SoundRecordingStoppedEvent())

        if os.path.exists(fifo):
            if stat.S_ISFIFO(os.stat(fifo).st_mode):
                self.logger.info('Removing previous input stream FIFO {}'.format(fifo))
                os.unlink(fifo)
            else:
                raise RuntimeError('{} exists and is not a FIFO. Please remove it or rename it'.format(fifo))

        os.mkfifo(fifo, 0o644)
        Thread(target=streaming_thread).start()

    @action
    def record(self, outfile=None, duration=None, device=None, sample_rate=None,
               format=None, blocksize=None, latency=0, channels=1, subtype='PCM_24'):
        """
        Records audio to a sound file (support formats: wav, raw)

        :param outfile: Sound file (default: the method will create a temporary file with the recording)
        :type outfile: str

        :param duration: Recording duration in seconds (default: record until stop event)
        :type duration: float

        :param device: Input device (default: default configured device or system default audio input if not configured)
        :type device: int or str

        :param sample_rate: Recording sample rate (default: device default rate)
        :type sample_rate: int

        :param format: Audio format (default: WAV)
        :type format: str

        :param blocksize: Audio block size (default: configured `input_blocksize` or 2048)
        :type blocksize: int

        :param latency: Device latency in seconds (default: 0)
        :type latency: float

        :param channels: Number of channels (default: 1)
        :type channels: int

        :param subtype: Recording subtype - see `Soundfile docs - Subtypes <https://pysoundfile.readthedocs.io/en/0.9.0/#soundfile.available_subtypes>`_ for a list of the available subtypes (default: PCM_24)
        :type subtype: str
        """

        def recording_thread(outfile, duration, device, sample_rate, format,
                             blocksize, latency, channels, subtype):
            import sounddevice as sd

            self.recording_paused_changed.clear()

            if outfile:
                outfile = os.path.abspath(os.path.expanduser(outfile))
                if os.path.isfile(outfile):
                    self.logger.info('Removing existing audio file {}'.format(outfile))
                    os.unlink(outfile)
            else:
                outfile = tempfile.NamedTemporaryFile(
                    prefix='recording_', suffix='.wav', delete=False,
                    dir=tempfile.gettempdir()).name

            if device is None:
                device = self.input_device
            if device is None:
                device = self._get_default_device('input')

            if sample_rate is None:
                dev_info = sd.query_devices(device, 'input')
                sample_rate = int(dev_info['default_samplerate'])

            if blocksize is None:
                blocksize = self.input_blocksize

            q = queue.Queue()

            def audio_callback(indata, frames, duration, status):
                while self._get_recording_state() == RecordingState.PAUSED:
                    self.recording_paused_changed.wait()

                if status:
                    self.logger.warning('Recording callback status: {}'.format(
                        str(status)))

                q.put({
                    'timestamp': time.time(),
                    'frames': frames,
                    'time': duration,
                    'data': indata.copy()
                })

            try:
                import soundfile as sf
                import numpy

                with sf.SoundFile(outfile, mode='w', samplerate=sample_rate,
                                  format=format, channels=channels, subtype=subtype) as f:
                    with sd.InputStream(samplerate=sample_rate, device=device,
                                        channels=channels, callback=audio_callback,
                                        latency=latency, blocksize=blocksize):
                        self.start_recording()
                        get_bus().post(SoundRecordingStartedEvent(filename=outfile))
                        self.logger.info('Started recording from device [{}] to [{}]'.
                                         format(device, outfile))

                        recording_started_time = time.time()

                        while self._get_recording_state() != RecordingState.STOPPED \
                                and (duration is None or
                                     time.time() - recording_started_time < duration):
                            while self._get_recording_state() == RecordingState.PAUSED:
                                self.recording_paused_changed.wait()

                            get_args = {
                                'block': True,
                                'timeout': max(0, duration - (time.time() - recording_started_time)),
                            } if duration is not None else {}

                            data = q.get(**get_args)
                            if data and time.time() - data.get('timestamp') <= 1.0:
                                # Only write the block if the latency is still acceptable
                                f.write(data['data'])

                    f.flush()

            except queue.Empty:
                self.logger.warning('Recording timeout: audio callback failed?')
            finally:
                self.stop_recording()
                get_bus().post(SoundRecordingStoppedEvent(filename=outfile))

        Thread(target=recording_thread,
               args=(
                   outfile, duration, device, sample_rate, format, blocksize, latency, channels, subtype)
               ).start()

    @action
    def recordplay(self, duration=None, input_device=None, output_device=None,
                   sample_rate=None, blocksize=None, latency=0, channels=1, dtype=None):
        """
        Records audio and plays it on an output sound device (audio pass-through)

        :param duration: Recording duration in seconds (default: record until stop event)
        :type duration: float

        :param input_device: Input device (default: default configured device or system default audio input if not configured)
        :type input_device: int or str

        :param output_device: Output device (default: default configured device or system default audio output if not configured)
        :type output_device: int or str

        :param sample_rate: Recording sample rate (default: device default rate)
        :type sample_rate: int

        :param blocksize: Audio block size (default: configured `output_blocksize` or 2048)
        :type blocksize: int

        :param latency: Device latency in seconds (default: 0)
        :type latency: float

        :param channels: Number of channels (default: 1)
        :type channels: int

        :param dtype: Data type for the recording - see `Soundfile docs - Recording <https://python-sounddevice.readthedocs.io/en/0.3.12/_modules/sounddevice.html#rec>`_ for available types (default: input device default)
        :type dtype: str
        """

        import sounddevice as sd

        self.recording_paused_changed.clear()

        if input_device is None:
            input_device = self.input_device
        if input_device is None:
            input_device = self._get_default_device('input')

        if output_device is None:
            output_device = self.output_device
        if output_device is None:
            output_device = self._get_default_device('output')

        if sample_rate is None:
            dev_info = sd.query_devices(input_device, 'input')
            sample_rate = int(dev_info['default_samplerate'])

        if blocksize is None:
            blocksize = self.output_blocksize

        # noinspection PyUnusedLocal
        def audio_callback(indata, outdata, frames, time, status):
            while self._get_recording_state() == RecordingState.PAUSED:
                self.recording_paused_changed.wait()

            if status:
                self.logger.warning('Recording callback status: {}'.format(
                    str(status)))

            outdata[:] = indata

        stream_index = None

        try:
            import soundfile as sf
            import numpy

            stream_index = self._allocate_stream_index()
            stream = sd.Stream(samplerate=sample_rate, channels=channels,
                               blocksize=blocksize, latency=latency,
                               device=(input_device, output_device),
                               dtype=dtype, callback=audio_callback)
            self.start_recording()
            self._start_playback(stream_index=stream_index,
                                 stream=stream)

            self.logger.info('Started recording pass-through from device ' +
                             '[{}] to sound device [{}]'.
                             format(input_device, output_device))

            recording_started_time = time.time()

            while self._get_recording_state() != RecordingState.STOPPED \
                    and (duration is None or
                         time.time() - recording_started_time < duration):
                while self._get_recording_state() == RecordingState.PAUSED:
                    self.recording_paused_changed.wait()

                time.sleep(0.1)

        except queue.Empty as e:
            self.logger.warning('Recording timeout: audio callback failed?')
        finally:
            self.stop_playback([stream_index])
            self.stop_recording()

    @action
    def query_streams(self):
        """
        :returns: A list of active audio streams
        """

        streams = {
            i: {
                attr: getattr(stream, attr)
                for attr in ['active', 'closed', 'stopped', 'blocksize',
                             'channels', 'cpu_load', 'device', 'dtype',
                             'latency', 'samplerate', 'samplesize']
                if hasattr(stream, attr)
            } for i, stream in self.active_streams.items()
        }

        for i, stream in streams.items():
            stream['playback_state'] = self.playback_state[i].name
            stream['name'] = self.stream_index_to_name.get(i)
            if i in self.stream_mixes:
                stream['mix'] = {j: sound for j, sound in
                                 enumerate(list(self.stream_mixes[i]))}

        return streams

    def _get_or_allocate_stream_index(self, stream_index=None, stream_name=None,
                                      completed_callback_event=None):
        stream = None

        with self.playback_state_lock:
            if stream_index is None:
                if stream_name is not None:
                    stream_index = self.stream_name_to_index.get(stream_name)
            else:
                if stream_name is not None:
                    raise RuntimeError('Redundant specification of both ' +
                                       'stream_name and stream_index')

            if stream_index is not None:
                stream = self.active_streams.get(stream_index)

            if not stream:
                return (self._allocate_stream_index(stream_name=stream_name,
                                                    completed_callback_event=
                                                    completed_callback_event),
                        True)

            return (stream_index, False)

    def _allocate_stream_index(self, stream_name=None,
                               completed_callback_event=None):
        stream_index = None

        with self.playback_state_lock:
            for i in range(len(self.active_streams) + 1):
                if i not in self.active_streams:
                    stream_index = i
                    break

            if stream_index is None:
                raise RuntimeError('No stream index available')

            if stream_name is None:
                stream_name = self._STREAM_NAME_PREFIX + str(stream_index)

            self.active_streams[stream_index] = None
            self.stream_mixes[stream_index] = Mix()
            self.stream_index_to_name[stream_index] = stream_name
            self.stream_name_to_index[stream_name] = stream_index
            self.completed_callback_events[stream_index] = \
                completed_callback_event if completed_callback_event else Event()

        return stream_index

    def _start_playback(self, stream_index, stream):
        with self.playback_state_lock:
            self.playback_state[stream_index] = PlaybackState.PLAYING
            self.active_streams[stream_index] = stream

            if isinstance(self.playback_paused_changed.get(stream_index), Event):
                self.playback_paused_changed[stream_index].clear()
            else:
                self.playback_paused_changed[stream_index] = Event()

        self.logger.info('Playback started on stream index {}'.
                         format(stream_index))

        return stream_index

    @action
    def stop_playback(self, streams=None):
        """
        :param streams: Streams to stop by index or name (default: all)
        :type streams: list[int] or list[str]
        """

        with self.playback_state_lock:
            streams = streams or self.active_streams.keys()
            if not streams:
                return
            completed_callback_events = {}

            for i in streams:
                stream = self.active_streams.get(i)
                if not stream:
                    i = self.stream_name_to_index.get(i)
                    stream = self.active_streams.get(i)
                if not stream:
                    self.logger.info('No such stream index or name: {}'.
                                     format(i))
                    continue

                if self.completed_callback_events[i]:
                    completed_callback_events[i] = self.completed_callback_events[i]
                self.playback_state[i] = PlaybackState.STOPPED

        for i, event in completed_callback_events.items():
            event.wait()

            if i in self.completed_callback_events:
                del self.completed_callback_events[i]
            if i in self.active_streams:
                del self.active_streams[i]
            if i in self.stream_mixes:
                del self.stream_mixes[i]

            if i in self.stream_index_to_name:
                name = self.stream_index_to_name[i]
                del self.stream_index_to_name[i]
                if name in self.stream_name_to_index:
                    del self.stream_name_to_index[name]

        self.logger.info('Playback stopped on streams [{}]'.format(
            ', '.join([str(stream) for stream in
                       completed_callback_events.keys()])))

    @action
    def pause_playback(self, streams=None):
        """
        :param streams: Streams to pause by index (default: all)
        :type streams: list[int]
        """

        with self.playback_state_lock:
            streams = streams or self.active_streams.keys()
            if not streams:
                return

            for i in streams:
                stream = self.active_streams.get(i)
                if not stream:
                    i = self.stream_name_to_index.get(i)
                    stream = self.active_streams.get(i)
                if not stream:
                    self.logger.info('No such stream index or name: {}'.
                                     format(i))
                    continue

                stream = self.active_streams[i]
                if self.playback_state[i] == PlaybackState.PAUSED:
                    self.playback_state[i] = PlaybackState.PLAYING
                elif self.playback_state[i] == PlaybackState.PLAYING:
                    self.playback_state[i] = PlaybackState.PAUSED
                else:
                    continue

                self.playback_paused_changed[i].set()

        self.logger.info('Playback pause toggled on streams [{}]'.format(
            ', '.join([str(stream) for stream in streams])))

    def start_recording(self):
        with self.recording_state_lock:
            self.recording_state = RecordingState.RECORDING

    @action
    def stop_recording(self):
        with self.recording_state_lock:
            self.recording_state = RecordingState.STOPPED
        self.logger.info('Recording stopped')

    @action
    def pause_recording(self):
        with self.recording_state_lock:
            if self.recording_state == RecordingState.PAUSED:
                self.recording_state = RecordingState.RECORDING
            elif self.recording_state == RecordingState.RECORDING:
                self.recording_state = RecordingState.PAUSED
            else:
                return

        self.logger.info('Recording paused state toggled')
        self.recording_paused_changed.set()

    @action
    def release(self, stream_index=None, stream_name=None,
                sound_index=None, midi_note=None, frequency=None):
        """
        Remove a sound from an active stream, either by sound index (use
            :meth:`platypush.sound.plugin.SoundPlugin.query_streams` to get
            the sounds playing on the active streams), midi_note, frequency
            or absolute file path.

        :param stream_index: Stream index (default: sound removed from all the
            active streams)
        :type stream_index: int

        :param stream_name: Stream name (default: sound removed from all the
            active streams)
        :type stream_index: str

        :param sound_index: Sound index
        :type sound_index: int

        :param midi_note: MIDI note
        :type midi_note: int

        :param frequency: Sound frequency
        :type frequency: float
        """

        if stream_name:
            if stream_index:
                raise RuntimeError('stream_index and stream name are ' +
                                   'mutually exclusive')
            stream_index = self.stream_name_to_index.get(stream_name)

        mixes = {
            i: mix for i, mix in self.stream_mixes.items()
        } if stream_index is None else {
            stream_index: self.stream_mixes[stream_index]
        }

        streams_to_stop = []

        for i, mix in mixes.items():
            for j, sound in enumerate(mix):
                if (sound_index is not None and j == sound_index) or \
                        (midi_note is not None
                         and sound.get('midi_note') == midi_note) or \
                        (frequency is not None
                         and sound.get('frequency') == frequency):
                    if len(list(mix)) == 1:
                        # Last sound in the mix
                        streams_to_stop.append(i)
                    else:
                        mix.remove(j)

        if streams_to_stop:
            self.stop_playback(streams_to_stop)

    def _get_playback_state(self, stream_index):
        with self.playback_state_lock:
            return self.playback_state[stream_index]

    def _get_recording_state(self):
        with self.recording_state_lock:
            return self.recording_state


# vim:sw=4:ts=4:et:



import todoist.models

import datetime

from typing import Optional, List, Dict, Any

from platypush.message import Mapping
from platypush.message.response import Response


class TodoistResponse(Response):
    pass


class TodoistUserResponse(TodoistResponse):
    def __init__(self,
                 auto_reminder: Optional[int] = None,
                 avatar_big: Optional[str] = None,
                 avatar_medium: Optional[str] = None,
                 avatar_s640: Optional[str] = None,
                 avatar_small: Optional[str] = None,
                 business_account_id: Optional[int] = None,
                 daily_goal: Optional[int] = None,
                 date_format: Optional[str] = None,
                 dateist_inline_disabled: Optional[bool] = None,
                 dateist_lang: Optional[str] = None,
                 days_off: Optional[List[int]] = None,
                 default_reminder: Optional[str] = None,
                 email: Optional[str] = None,
                 features: Optional[Dict[str, Any]] = None,
                 full_name: Optional[str] = None,
                 id: Optional[int] = None,
                 image_id: Optional[str] = None,
                 inbox_project: Optional[int] = None,
                 is_biz_admin: Optional[bool] = None,
                 is_premium: Optional[bool] = None,
                 join_date: Optional[datetime.datetime] = None,
                 karma: Optional[float] = None,
                 karma_trend: Optional[str] = None,
                 lang: Optional[str] = None,
                 legacy_inbox_project: Optional[int] = None,
                 mobile_host: Optional[str] = None,
                 mobile_number: Optional[str] = None,
                 next_week: Optional[int] = None,
                 premium_until: Optional[datetime.datetime] = None,
                 share_limit: Optional[int] = None,
                 sort_order: Optional[int] = None,
                 start_day: Optional[int] = None,
                 start_page: Optional[str] = None,
                 theme: Optional[int] = None,
                 time_format: Optional[int] = None,
                 token: Optional[str] = None,
                 tz_info: Optional[Dict[str, Any]] = None,
                 unique_prefix: Optional[int] = None,
                 websocket_url: Optional[str] = None,
                 weekly_goal: Optional[int] = None,
                 **kwargs):
        response = {
            'auto_reminder': auto_reminder,
            'avatar_big': avatar_big,
            'avatar_medium': avatar_medium,
            'avatar_s640': avatar_s640,
            'avatar_small': avatar_small,
            'business_account_id': business_account_id,
            'daily_goal': daily_goal,
            'date_format': date_format,
            'dateist_inline_disabled': dateist_inline_disabled,
            'dateist_lang': dateist_lang,
            'days_off': days_off,
            'default_reminder': default_reminder,
            'email': email,
            'features': features,
            'full_name': full_name,
            'id': id,
            'image_id': image_id,
            'inbox_project': inbox_project,
            'is_biz_admin': is_biz_admin,
            'is_premium': is_premium,
            'join_date': join_date,
            'karma': karma,
            'karma_trend': karma_trend,
            'lang': lang,
            'legacy_inbox_project': legacy_inbox_project,
            'mobile_host': mobile_host,
            'mobile_number': mobile_number,
            'next_week': next_week,
            'premium_until': premium_until,
            'share_limit': share_limit,
            'sort_order': sort_order,
            'start_day': start_day,
            'start_page': start_page,
            'theme': theme,
            'time_format': time_format,
            'token': token,
            'tz_info': tz_info,
            'unique_prefix': unique_prefix,
            'websocket_url': websocket_url,
            'weekly_goal': weekly_goal,
        }

        super().__init__(output=response, **kwargs)


class TodoistProject(Mapping):
    def __init__(self,
                 child_order: int,
                 collapsed: int,
                 color: int,
                 has_more_notes: bool,
                 id: int,
                 is_archived: bool,
                 is_deleted: bool,
                 is_favorite: bool,
                 name: str,
                 shared: bool,
                 inbox_project: Optional[bool] = None,
                 legacy_id: Optional[int] = None,
                 parent_id: Optional[int] = None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.child_order = child_order
        self.collapsed = collapsed
        self.color = color
        self.has_more_notes = has_more_notes
        self.id = id
        self.inbox_project = inbox_project
        self.is_archived = bool(is_archived)
        self.is_deleted = bool(is_deleted)
        self.is_favorite = bool(is_favorite)
        self.name = name
        self.shared = shared
        self.legacy_id = legacy_id
        self.parent_id = parent_id


class TodoistProjectsResponse(TodoistResponse):
    def __init__(self, projects: List[TodoistProject], **kwargs):
        self.projects = [TodoistProject(**(p.data if isinstance(p, todoist.models.Project) else p)) for p in projects]
        super().__init__(output=[p.__dict__ for p in self.projects], **kwargs)


class TodoistItem(Mapping):
    def __init__(self,
                 content: str,
                 id: int,
                 checked: bool,
                 priority: int,
                 child_order: int,
                 collapsed: bool,
                 day_order: int,
                 date_added: datetime.datetime,
                 in_history: bool,
                 is_deleted: bool,
                 user_id: int,
                 has_more_notes: bool = False,
                 project_id: Optional[int] = None,
                 parent_id: Optional[int] = None,
                 responsible_uid: Optional[int] = None,
                 date_completed: Optional[datetime.datetime] = None,
                 assigned_by_uid: Optional[int] = None,
                 due: Optional[Dict[str, Any]] = None,
                 labels: Optional[List[str]] = None,
                 legacy_project_id: Optional[int] = None,
                 section_id: Optional[int] = None,
                 sync_id: Optional[int] = None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.content = content
        self.id = id
        self.checked = bool(checked)
        self.priority = priority
        self.child_order = child_order
        self.collapsed = bool(collapsed)
        self.day_order = day_order
        self.date_added = date_added
        self.has_more_notes = bool(has_more_notes)
        self.in_history = bool(in_history)
        self.is_deleted = bool(is_deleted)
        self.user_id = user_id
        self.project_id = project_id
        self.parent_id = parent_id
        self.responsible_uid = responsible_uid
        self.date_completed = date_completed
        self.assigned_by_uid = assigned_by_uid
        self.due = due
        self.labels = labels
        self.legacy_project_id = legacy_project_id
        self.section_id = section_id
        self.sync_id = sync_id


class TodoistItemsResponse(TodoistResponse):
    def __init__(self, items: List[TodoistItem], **kwargs):
        self.items = [TodoistItem(**(i.data if isinstance(i, todoist.models.Item) else i.__dict__)) for i in items]
        super().__init__(output=[i.__dict__ for i in self.items], **kwargs)


class TodoistFilter(Mapping):
    def __init__(self,
                 color: [int],
                 id: [int],
                 is_deleted: [bool],
                 is_favorite: [bool],
                 item_order: [int],
                 name: [str],
                 query: [str],
                 legacy_id: Optional[int] = None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.color = color
        self.id = id
        self.is_deleted = is_deleted
        self.is_favorite = is_favorite
        self.item_order = item_order
        self.name = name
        self.query = query
        self.legacy_id = legacy_id


class TodoistFiltersResponse(TodoistResponse):
    def __init__(self, filters: List[TodoistFilter], **kwargs):
        self.filters = [TodoistFilter(**(f.data if isinstance(f, todoist.models.Filter) else f.__dict__))
                        for f in filters]

        super().__init__(output=[f.__dict__ for f in self.filters], **kwargs)


class TodoistLiveNotification(Mapping):
    def __init__(self,
                 id: [int],
                 is_deleted: [bool],
                 created: [str],
                 is_unread: [bool],
                 notification_key: [str],
                 notification_type: [str],
                 completed_last_month: Optional[int] = None,
                 karma_level: Optional[int] = None,
                 promo_img: Optional[str] = None,
                 completed_tasks: Optional[int] = None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.id = id
        self.is_deleted = bool(is_deleted)
        self.completed_last_month = completed_last_month
        self.completed_tasks = completed_tasks
        self.created = created
        self.is_unread = bool(is_unread)
        self.karma_level = karma_level
        self.notification_key = notification_key
        self.notification_type = notification_type
        self.promo_img = promo_img


class TodoistLiveNotificationsResponse(TodoistResponse):
    def __init__(self, notifications: List[TodoistLiveNotification], **kwargs):
        self.notifications = [TodoistLiveNotification(**(n.data if isinstance(n, todoist.models.LiveNotification)
                                                         else n.__dict__)) for n in notifications]

        super().__init__(output=[n.__dict__ for n in self.notifications], **kwargs)


class TodoistCollaborator(Mapping):
    def __init__(self, data: Dict[str, Any], *args, **kwargs):
        super().__init__(*args, **kwargs)
        for k, v in data.items():
            self.__setattr__(k, v)


class TodoistCollaboratorsResponse(TodoistResponse):
    def __init__(self, collaborators: List[TodoistCollaborator], **kwargs):
        self.collaborators = [TodoistCollaborator(c.data if isinstance(c, todoist.models.Collaborator) else c.__dict__)
                              for c in collaborators]

        super().__init__(output=[c.__dict__ for c in self.collaborators], **kwargs)


class TodoistNote(Mapping):
    def __init__(self, data: Dict[str, Any], *args, **kwargs):
        super().__init__(*args, **kwargs)
        for k, v in data.items():
            self.__setattr__(k, v)


class TodoistNotesResponse(TodoistResponse):
    def __init__(self, notes: List[TodoistCollaborator], **kwargs):
        self.notes = [TodoistCollaborator(n.data if isinstance(n, todoist.models.Note) else n.__dict__)
                      for n in notes]

        super().__init__(output=[n.__dict__ for n in self.notes], **kwargs)


class TodoistProjectNote(Mapping):
    def __init__(self, data: Dict[str, Any], *args, **kwargs):
        super().__init__(*args, **kwargs)
        for k, v in data.items():
            self.__setattr__(k, v)


class TodoistProjectNotesResponse(TodoistResponse):
    def __init__(self, notes: List[TodoistCollaborator], **kwargs):
        self.notes = [TodoistCollaborator(n.data if isinstance(n, todoist.models.ProjectNote) else n.__dict__)
                      for n in notes]

        super().__init__(output=[n.__dict__ for n in self.notes], **kwargs)


class TodoistReminder(Mapping):
    def __init__(self, data: Dict[str, Any], *args, **kwargs):
        super().__init__(*args, **kwargs)
        for k, v in data.items():
            self.__setattr__(k, v)


class TodoistRemindersResponse(TodoistResponse):
    def __init__(self, reminders: List[TodoistReminder], **kwargs):
        self.reminders = [TodoistReminder(n.data if isinstance(n, todoist.models.Reminder) else n.__dict__)
                          for n in reminders]

        super().__init__(output=[r.__dict__ for r in self.reminders], **kwargs)


# vim:sw=4:ts=4:et:



from platypush.message.event import Event


class ConnectedEvent(Event):
    """
    Event triggered when the backend connects to the Adafruit message queue
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, feed=feed, data=data, **kwargs)


class DisconnectedEvent(Event):
    """
    Event triggered when the backend disconnects from the Adafruit message queue
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, feed=feed, data=data, **kwargs)


class FeedUpdateEvent(Event):
    """
    Event triggered upon Adafruit IO feed update
    """

    def __init__(self, feed, data, *args, **kwargs):
        super().__init__(*args, feed=feed, data=data, **kwargs)


# vim:sw=4:ts=4:et:



import datetime
import re

from flask import Blueprint, request, redirect, render_template, make_response

from platypush.backend.http.app import template_folder
from platypush.backend.http.utils import HttpUtils
from platypush.user import UserManager

login = Blueprint('login', __name__, template_folder=template_folder)

# Declare routes list
__routes__ = [
    login,
]


@login.route('/login', methods=['GET', 'POST'])
def login():
    """ Login page """
    user_manager = UserManager()
    session_token = request.cookies.get('session_token')

    # redirect_page = request.args.get('redirect', request.headers.get('Referer', '/'))
    redirect_page = request.args.get('redirect')
    if not redirect_page:
        redirect_page = request.headers.get('Referer', '/')
    if re.search('(^https?://[^/]+)?/login[^?#]?', redirect_page):
        # Prevent redirect loop
        redirect_page = '/'

    if session_token:
        user, session = user_manager.authenticate_user_session(session_token)
        if user:
            return redirect(redirect_page, 302)

    if request.form:
        username = request.form.get('username')
        password = request.form.get('password')
        remember = request.form.get('remember')
        expires = datetime.datetime.utcnow() + datetime.timedelta(days=365) \
            if remember else None

        session = user_manager.create_user_session(username=username, password=password,
                                                   expires_at=expires)

        if session:
            redirect_target = redirect(redirect_page, 302)
            response = make_response(redirect_target)
            response.set_cookie('session_token', session.session_token, expires=expires)
            return response

    return render_template('login.html', utils=HttpUtils)


# vim:sw=4:ts=4:et:



from flask import Blueprint, request, render_template

from platypush.backend.http.app import template_folder, static_folder
from platypush.backend.http.app.utils import authenticate, get_websocket_port

from platypush.backend.http.utils import HttpUtils
from platypush.config import Config

dashboard = Blueprint('dashboard', __name__, template_folder=template_folder)

# Declare routes list
__routes__ = [
    dashboard,
]


@dashboard.route('/dashboard', methods=['GET'])
@authenticate()
def dashboard():
    """ Route for the fullscreen dashboard """
    http_conf = Config.get('backend.http')
    dashboard_conf = http_conf.get('dashboard', {})

    return render_template('dashboard.html', config=dashboard_conf,
                           utils=HttpUtils, token=Config.get('token'),
                           static_folder=static_folder, template_folder=template_folder,
                           websocket_port=get_websocket_port(),
                           has_ssl=http_conf.get('ssl_cert') is not None)


# vim:sw=4:ts=4:et:



import base64
import io
import os
import threading

from typing import Dict, Optional, List, Any, Union

from platypush import Response
from platypush.message.response.esp import EspWifiScanResult, EspWifiConfigResult
from platypush.plugins import Plugin, action
from platypush.plugins.esp.models.connection import Connection
from platypush.plugins.esp.models.device import Device


class EspPlugin(Plugin):
    # noinspection PyUnresolvedReferences
    """
        This plugin allows you to fully control to ESP8266/ESP32 devices connected over WiFi.
        It uses the WebREPL interface embedded in MicroPython to communicate with the device.

        All you need to do is to flash the MicroPython firmware to your device, enable the WebREPL interface,
        and you can use this plugin to fully control the device remotely without deploying any code to the controller.

            - Download the `MicroPython firmware <https://micropython.org/download>`_ for your device.
            - Connect your ESP8266/ESP32 via serial/USB port and
              `flash the firmware <https://docs.micropython.org/en/latest/esp8266/tutorial/intro.html#intro>`_.
              For example. using ``esptool`` and assuming that you have an ESP8266 device connected on ``/dev/ttyUSB0``:

              .. code-block:: shell

                  # Erase the flash memory
                  esptool.py --port /dev/ttyUSB0 erase_flash
                  # Flash the firmware
                  esptool.py --port /dev/ttyUSB0 --baud 115200 write_flash --flash_size=detect 0 esp8266-[version].bin

            - Access the MicroPython interpreter over serial/USB port. For example, on Linux:

            .. code-block:: shell

                picocom /dev/ttyUSB0 -b11520

            - Configure the WiFi interface:

            .. code-block:: python

                >>> import network
                >>> wlan = network.WLAN(network.STA_IF)
                >>> wlan.active(True)
                >>> wlan.connect('YourSSID', 'YourPassword')
                >>> # Print the device IP address
                >>> wlan.ifconfig()[0]
                >>> '192.168.1.23'

            - Enable the`WebREPL <https://docs.micropython.org/en/latest/esp8266/quickref.html#webrepl-web-browser-interactive-prompt>`_
              interface on the device:

              .. code-block:: python

                  >>> import webrepl_setup

            - Follow the instructions, set a password and reset your device. A websocket service should be available
              by default on the port 8266 of your ESP8266/ESP32 and it can accept commands sent by platypush.

        Requires:

            * **websocket-client** (``pip install websocket-client``)

        """

    def __init__(self, devices: List[Union[Device, dict]] = None, **kwargs):
        """
        :param devices: List of configured device. Pre-configuring devices by name allows you to call the actions
            directly by device name, instead of specifying ``host``, ``port`` and ``password`` on each call. It
            also allows you to interact with PINs by name, if you specified names for them, instead of using the
            PIN number on your calls. Example configuration:

            .. code-block:: yaml

                devices:
                    - host: 192.168.1.23
                      port: 8266        # WebREPL websocket port
                      password: pwd1    # Device password
                      name: smoke_detector
                      pins:
                        - number: 14
                          name: smoke_sensor
                          pwm: False

                    - host: 192.168.1.24
                      port: 8266
                      password: pwd2
                      name: smart_switch
                      pins:
                        - number: 13
                          name: relay
                          pwm: True

        """

        super().__init__(**kwargs)
        self.devices = [
            dev if isinstance(dev, Device) else Device(**dev)
            for dev in (devices or [])
        ]

        self._devices_by_host = {dev['host']: dev for dev in self.devices}
        self._devices_by_name = {dev['name']: dev for dev in self.devices if dev['name']}
        self._connections: Dict[tuple, Connection] = {}

    def __del__(self):
        if not self._connections:
            return
        for conn in self._connections.values():
            conn.close()

    def on_open(self, conn: Connection):
        def callback(ws):
            conn.on_connect()
            self.logger.info('Connection to {} opened'.format(ws.url))

        return callback

    def on_message(self, conn: Connection):
        def handler(ws, msg):
            if not isinstance(msg, str):
                # Bytes sequences will be handled by on_data
                return

            if msg.endswith('Password: ') and conn.state == conn.State.CONNECTED:
                conn.on_password_requested()
                return

            if conn.state in (conn.State.CONNECTED, conn.state.PASSWORD_REQUIRED) and msg.endswith('>>> '):
                conn.on_ready()
                return

            if conn.state == conn.State.WAITING_ECHO:
                conn.on_recv_echo(msg)
                return

            if conn.state == conn.State.WAITING_RESPONSE:
                if msg.endswith('>>> '):
                    msg = msg[:-4]
                    conn.on_recv_response(msg)
                else:
                    conn.append_response(msg)

                return

            self.logger.debug('Message received on {}: {}'.format(ws.url, msg))

        def callback(ws, msg):
            try:
                handler(ws, msg)
            except Exception as e:
                self.logger.exception(e)
                raise e

        return callback

    def on_data(self, conn: Connection):
        # noinspection PyUnusedLocal
        def handler(ws, data):
            if conn.state == conn.State.WAITING_FILE_TRANSFER_RESPONSE:
                conn.on_recv_file_transfer_response(data)
                return

            if conn.state == conn.State.UPLOADING_FILE:
                conn.on_file_transfer_completed(data)
                return

            if conn.state == conn.State.DOWNLOADING_FILE:
                conn.on_chunk_received(data)
                return

        # noinspection PyUnusedLocal
        def callback(ws, data, data_type, continue_flag):
            import websocket

            try:
                if data_type == websocket.ABNF.OPCODE_BINARY:
                    handler(ws, data)
            except Exception as e:
                self.logger.exception(e)
                raise e

        return callback

    def on_close(self, conn: Connection):
        def callback(ws):
            # noinspection PyBroadException
            try:
                ws.close()
            except:
                pass

            conn.on_close()
            self.logger.info('Connection to {}:{} closed'.format(conn.host, conn.port))

        return callback

    def on_error(self, conn: Connection):
        def callback(ws, err):
            conn.on_close()
            self.logger.warning('Error on the connection to {}: {}'.format(ws.url, err))

        return callback

    # noinspection PyUnusedLocal
    def _get_device(self, device: Optional[str] = None, host: Optional[str] = None, port: int = 8266,
                    password: Optional[str] = None, **kwargs) -> Device:
        if device:
            assert device in self._devices_by_name, 'No such device configured: ' + device
            return self._devices_by_name[device]

        assert host and port, 'No host and port specified'
        if host in self._devices_by_host:
            return self._devices_by_host[host]

        return Device(host=host, port=port, password=password)

    # noinspection PyUnusedLocal
    def _get_connection(self, device: Optional[str] = None, host: Optional[str] = None, port: int = 8266, **kwargs) \
            -> Connection:
        if device:
            assert device in self._devices_by_name, 'No such device configured: ' + device
            device = self._devices_by_name[device]
            host = device['host']
            port = device['port']

        assert host and port, 'No host and port specified'
        return self._connections.get((host, port))

    @action
    def connect(self, device: Optional[str] = None, host: Optional[str] = None, port: int = 8266,
                password: Optional[str] = None, timeout: Optional[float] = 10.0):
        """
        Open a connection to an ESP device.

        :param device: Configured device name. Either ``device`` or ``host`` and ``port`` must be provided.
        :param host: ESP host.
        :param port: ESP port (default: 8266).
        :param password: ESP WebREPL password.
        :param timeout: Connection timeout (default: 10 seconds).
        """
        import websocket

        device = self._get_device(device=device, host=host, port=port, password=password)
        host = device['host']
        port = device['port']
        conn = self._get_connection(host=host, port=port)
        if conn and conn.ws and conn.ws.sock.connected:
            self.logger.info('Already connected to {}:{}'.format(host, port))
            return

        conn = Connection(host=host, port=port, password=password, connect_timeout=timeout)
        self._connections[(host, port)] = conn

        ws = websocket.WebSocketApp('ws://{host}:{port}'.format(host=host, port=port),
                                    on_open=self.on_open(conn),
                                    on_message=self.on_message(conn),
                                    on_data=self.on_data(conn),
                                    on_error=self.on_error(conn),
                                    on_close=self.on_close(conn))

        conn.ws = ws
        wst = threading.Thread(target=ws.run_forever)
        wst.daemon = True
        wst.start()
        conn.wait_ready()

    @action
    def close(self, device: Optional[str] = None, host: Optional[str] = None, port: int = 8266):
        """
        Close an active connection to a device.
        :param device: Configured device name. Either ``device`` or ``host`` and ``port`` must be provided.
        :param host: ESP host.
        :param port: ESP port.
        """
        device = self._get_device(device=device, host=host, port=port)
        host, port = [device['host'], device['port']]
        conn = self._connections.get((host, port))
        assert conn, 'No active connection found to {}:{}'.format(host, port)
        conn.close()
        del self._connections[(host, port)]

    # noinspection PyUnusedLocal
    @action
    def execute(self,
                code: str,
                device: Optional[str] = None,
                host: Optional[str] = None,
                port: int = 8266,
                password: Optional[str] = None,
                conn_timeout: Optional[float] = 10.0,
                recv_timeout: Optional[float] = 30.0,
                wait_response: bool = True,
                **kwargs) -> Response:
        """
        Run raw Python code on the ESP device.

        :param code: Snippet of code to run.
        :param device: Configured device name. Either ``device`` or ``host`` and ``port`` must be provided.
        :param host: ESP host.
        :param port: ESP port (default: 8266).
        :param password: ESP WebREPL password.
        :param conn_timeout: Connection timeout (default: 10 seconds).
        :param recv_timeout: Response receive timeout (default: 30 seconds).
        :param wait_response: Wait for the response from the device (default: True)
        :return: The response returned by the Micropython interpreter, as a string.
        """
        device = self._get_device(device=device, host=host, port=port, password=password)
        self.connect(host=device['host'], port=device['port'], password=device['password'], timeout=conn_timeout)
        conn = self._connections.get((device['host'], device['port']))

        try:
            return conn.send(code, timeout=recv_timeout, wait_response=wait_response)
        except Exception as e:
            conn.close()
            raise e

    @action
    def pin_on(self, pin: Union[int, str], pull_up: bool = False, **kwargs):
        """
        Set the specified PIN to HIGH.

        :param pin: GPIO PIN number or configured name.
        :param pull_up: Set to True if the PIN has a (weak) pull-up resistor attached.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        device = self._get_device(**kwargs)
        pin = device.get_pin(pin)
        code = '''
import machine
pin = machine.Pin({pin}, machine.Pin.OUT{pull_up})
pin.on()
'''.format(pin=pin, pull_up=', machine.Pin.PULL_UP' if pull_up else '')

        self.execute(code, **kwargs)

    @action
    def pin_off(self, pin: Union[int, str], pull_up: bool = False, **kwargs):
        """
        Set the specified PIN to LOW.

        :param pin: GPIO PIN number.
        :param pull_up: Set to True if the PIN has a (weak) pull-up resistor attached.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        device = self._get_device(**kwargs)
        pin = device.get_pin(pin)
        code = '''
import machine
pin = machine.Pin({pin}, machine.Pin.OUT{pull_up})
pin.off()
'''.format(pin=pin, pull_up=', machine.Pin.PULL_UP' if pull_up else '')

        self.execute(code, **kwargs)

    @action
    def pin_toggle(self, pin: Union[int, str], pull_up: bool = False, **kwargs):
        """
        Toggle a PIN state - to HIGH if LOW, to LOW if HIGH.

        :param pin: GPIO PIN number or configured name.
        :param pull_up: Set to True if the PIN has a (weak) pull-up resistor attached.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        device = self._get_device(**kwargs)
        pin = device.get_pin(pin)
        code = '''
import machine
pin = machine.Pin({pin}, machine.Pin.OUT{pull_up})
if pin.value():
    pin.off()
else:
    pin.on()
'''.format(pin=pin, pull_up=', machine.Pin.PULL_UP' if pull_up else '')

        self.execute(code, **kwargs)

    @action
    def pin_read(self, pin: Union[int, str], out: bool = False, pull_up: bool = False, **kwargs) -> bool:
        """
        Get the ON/OFF value of a PIN.

        :param pin: GPIO PIN number or configured name.
        :param out: Treat the PIN as an output PIN - e.g. if you usually write to it and now want to read the
            value. If not set, then the PIN will be treated as an input PIN.
        :param pull_up: Set to True if the PIN has a (weak) pull-up resistor attached.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        device = self._get_device(**kwargs)
        pin = device.get_pin(pin)
        code = '''
import machine
pin = machine.Pin({pin}, machine.Pin.{inout}{pull_up})
pin.value()
'''.format(pin=pin, inout='OUT' if out else 'IN', pull_up=', machine.Pin.PULL_UP' if pull_up else '')

        return bool(self.execute(code, **kwargs).output)

    @action
    def adc_read(self, pin: int = 0, **kwargs) -> int:
        """
        Read an analog value from a PIN. Note that the ESP8266 only has one analog PIN, accessible on
        the channel ``0``. If you are interested in the actual voltage that is measured then apply
        ``V = Vcc * (value/1024)``, where ``Vcc`` is the supply voltage provided to the device (usually 3V if
        connected to the Vcc PIN of an ESP8266).

        :param pin: GPIO PIN number (default: 0).
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        :return: A value between ``0`` and ``1024``.
        """
        code = '''
import machine
adc = machine.ADC({pin})
adc.read()
'''.format(pin=pin)

        response = self.execute(code, **kwargs)
        return int(response.output)

    @action
    def pwm_freq(self, pin: Union[int, str], freq: Optional[int] = None, **kwargs) -> Optional[int]:
        """
        Get/set the frequency of a PWM PIN.

        :param pin: GPIO PIN number or configured name.
        :param freq: If set, set the frequency for the PIN in Hz.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        device = self._get_device(**kwargs)
        pin = device.get_pin(pin)
        code = '''
import machine
pin = machine.PWM(machine.Pin({pin}))
pin.freq({freq})
'''.format(pin=pin, freq=freq if freq else '')

        ret = self.execute(code, **kwargs).output
        if not freq:
            return int(ret)

    @action
    def pwm_duty(self, pin: Union[int, str], duty: Optional[int] = None, **kwargs) -> Optional[int]:
        """
        Get/set the duty cycle of a PWM PIN.

        :param pin: GPIO PIN number or configured name.
        :param duty: Optional duty value to set.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        device = self._get_device(**kwargs)
        pin = device.get_pin(pin)
        code = '''
import machine
pin = machine.PWM(machine.Pin({pin}))
pin.duty({duty})
'''.format(pin=pin, duty=duty if duty else '')

        ret = self.execute(code, **kwargs).output
        if not duty:
            return int(ret)

    @action
    def pwm_on(self, pin: Union[int, str], freq: Optional[int] = None, duty: Optional[int] = None, **kwargs):
        """
        Set the specified PIN to HIGH.

        :param pin: GPIO PIN number or configured name.
        :param freq: PWM PIN frequency.
        :param duty: PWM PIN duty cycle.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        device = self._get_device(**kwargs)
        pin = device.get_pin(pin)
        code = '''
import machine
pin = machine.PWM(machine.Pin({pin}))

if {freq}:
    pin.freq({freq})
if {duty}:
    pin.duty({duty})

pin.on()
'''.format(pin=pin, freq=freq, duty=duty)

        self.execute(code, **kwargs)

    @action
    def pwm_off(self, pin: Union[int, str], **kwargs):
        """
        Turn off a PWM PIN.

        :param pin: GPIO PIN number or configured name.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        device = self._get_device(**kwargs)
        pin = device.get_pin(pin)
        code = '''
import machine
pin = machine.PWM(machine.Pin({pin}))
pin.deinit()
'''.format(pin=pin)

        self.execute(code, **kwargs)

    @action
    def spi_open(self, id=1, baudrate: int = 1000000, polarity: int = 0, phase: int = 0,
                 bits: int = 8, sck: Optional[int] = None, mosi: Optional[int] = None,
                 miso: Optional[int] = None, **kwargs):
        """
        Open a connection to an SPI port.
        Note that ``sck``, ``mosi`` and ``miso`` parameters are only allowed if you're setting up a software
        managed SPI connection. If you're using the hardware SPI implementation then those PINs are
        pre-defined depending on the model of your board.

        :param id: Values of id depend on a particular port and its hardware. Values 0, 1, etc. are commonly used to
            select hardware SPI block #0, #1, etc. Value -1 can be used for bit-banging (software) implementation of
            SPI (if supported by a port).
        :param baudrate: Port baudrate/SCK clock rate (default: 1 MHz).
        :param polarity: It can be 0 or 1, and is the level the idle clock line sits at.
        :param phase: It can be 0 or 1 to sample data on the first or second clock edge respectively.
        :param bits: Number of bits per character. It can be 7, 8 or 9.
        :param sck: SCK PIN number.
        :param mosi: MOSI PIN number.
        :param miso: MISO PIN number.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        code = '''
args = {
    'baudrate': {baudrate},
    'polarity': {polarity},
    'phase': {phase},
    'bits': {bits},
}
'''.format(baudrate=baudrate, polarity=polarity, phase=phase, bits=bits)

        self.execute(code, **kwargs)
        code = '''
import machine

if {sck}:
    args['sck'] = machine.Pin({sck})
if {mosi}:
    args['mosi'] = machine.Pin({mosi})
if {miso}:
    args['miso'] = machine.Pin({miso})
'''.format(sck=sck, mosi=mosi, miso=miso)

        self.execute(code, **kwargs)
        code = 'spi = machine.SPI({id}, **args)'.format(id=id)
        self.execute(code, **kwargs)

    @action
    def spi_close(self, **kwargs):
        """
        Turn off an SPI bus.

        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.spi_open` and
            :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        self.spi_open(**kwargs)
        self.execute('spi.deinit()', **kwargs)

    @action
    def spi_read(self, size: int, **kwargs) -> str:
        """
        Read from an SPI bus.

        :param size: Number of bytes to read.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.spi_open` and
            :meth:`platypush.plugins.esp.EspPlugin.execute`.
        :return: String representation of the read bytes, or base64-encoded representation if the
            data can't be decoded to a string.
        """
        self.spi_open(**kwargs)
        code = 'spi.read({size})'.format(size=size)
        response = self.execute(code, **kwargs).output

        try:
            return response.decode()
        except UnicodeDecodeError:
            return base64.encodebytes(response).decode()

    @action
    def spi_write(self, data: str, binary: bool = False, **kwargs):
        """
        Write data to an SPI bus.

        :param data: Data to be written.
        :param binary: By default data will be treated as a string. Set binary to True if it should
            instead be treated as a base64-encoded binary string to be decoded before being sent.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.spi_open` and
            :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        if binary:
            data = base64.decodebytes(data.encode())
        else:
            data = data.encode()

        data = 'b"' + ''.join(['\\x{:02x}'.format(b) for b in data]) + '"'
        self.spi_open(**kwargs)

        code = 'spi.write({data})'.format(data=data)
        self.execute(code, **kwargs)

    @action
    def i2c_open(self, scl: Optional[int] = None, sda: Optional[int] = None, id: int = -1, baudrate: int = 400000, **kwargs):
        """
        Open a connection to an I2C (or "two-wire") port.

        :param scl: PIN number for the SCL (serial clock) line.
        :param sda: PIN number for the SDA (serial data) line.
        :param id: The default value of -1 selects a software implementation of I2C which can work (in most cases)
            with arbitrary pins for SCL and SDA. If id is -1 then scl and sda must be specified. Other allowed
            values for id depend on the particular port/board, and specifying scl and sda may or may not be required
            or allowed in this case.
        :param baudrate: Port frequency/clock rate (default: 400 kHz).
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        code = '''
import machine

args = {}
if {scl}:
    args['scl'] = machine.Pin(''' + str(scl) + ''')
if {sda}:
    args['sda'] = machine.Pin(''' + str(sda) + ''')
'''

        self.execute(code, **kwargs)

        code = '''
i2c = machine.I2C(id={id}, freq={baudrate}, **args)
'''.format(id=id, baudrate=baudrate)

        self.execute(code, **kwargs)

    @action
    def i2c_close(self, **kwargs):
        """
        Turn off an I2C bus.

        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.i2c_open` and
            :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        self.i2c_open(**kwargs)
        self.execute('i2c.deinit()', **kwargs)

    @action
    def i2c_scan(self, **kwargs) -> List[int]:
        """
        Scan for device addresses on the I2C bus.

        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.i2c_open` and
            :meth:`platypush.plugins.esp.EspPlugin.execute`.
        :return: List of 7-bit addresses.
        """
        self.i2c_open(**kwargs)
        return self.execute('i2c.scan', **kwargs).output

    @action
    def i2c_read(self, address: int, size: int, **kwargs) -> str:
        """
        Read data from the I2C bus.

        :param address: I2C address.
        :param size: Number of bytes to read.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.i2c_open` and
            :meth:`platypush.plugins.esp.EspPlugin.execute`.
        :return: String representation of the read bytes, or base64-encoded representation if the
            data can't be decoded to a string.
        """
        self.i2c_open(**kwargs)
        code = 'i2c.readfrom({address}, {size})'.format(address=address, size=size)
        response = self.execute(code, **kwargs).output

        try:
            return response.decode()
        except UnicodeDecodeError:
            return base64.encodebytes(response).decode()

    @action
    def i2c_write(self, address: int, data: str, binary: bool = False, **kwargs):
        """
        Write data to the I2C bus.

        :param address: I2C address.
        :param data: Data to be sent.
        :param binary: By default data will be treated as a string. Set binary to True if it should
            instead be treated as a base64-encoded binary string to be decoded before being sent.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.i2c_open` and
            :meth:`platypush.plugins.esp.EspPlugin.execute`.
        :return: String representation of the read bytes, or base64-encoded representation if the
            data can't be decoded to a string.
        """
        if binary:
            data = base64.decodebytes(data.encode())
        else:
            data = data.encode()

        data = 'b"' + ''.join(['\\x{:02x}'.format(b) for b in data]) + '"'
        self.i2c_open(**kwargs)
        code = 'i2c.writeto({address}, {data})'.format(address=address, data=data)
        self.execute(code, **kwargs)

    @action
    def i2c_start(self, **kwargs):
        """
        Generate a START condition on an I2C bus.

        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.i2c_open` and
            :meth:`platypush.plugins.esp.EspPlugin.execute`.
        :return: String representation of the read bytes, or base64-encoded representation if the
            data can't be decoded to a string.
        """
        self.i2c_open(**kwargs)
        self.execute('i2c.start()', **kwargs)

    @action
    def i2c_stop(self, **kwargs):
        """
        Generate a STOP condition on an I2C bus.

        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.i2c_open` and
            :meth:`platypush.plugins.esp.EspPlugin.execute`.
        :return: String representation of the read bytes, or base64-encoded representation if the
            data can't be decoded to a string.
        """
        self.i2c_open(**kwargs)
        self.execute('i2c.stop()', **kwargs)

    @action
    def uart_open(self, id=1, baudrate: Optional[int] = 9600, bits: Optional[int] = 8, parity: Optional[int] = None,
                  stop: int = 1, tx_pin: Optional[int] = None, rx_pin: Optional[int] = None,
                  timeout: Optional[float] = None, timeout_char: Optional[float] = None, **kwargs):
        """
        Open a connection to a UART port.

        :param id: Bus ID (default: 1).
        :param baudrate: Port baudrate (default: 9600).
        :param bits: Number of bits per character. It can be 7, 8 or 9.
        :param parity: Parity configuration. It can be None (no parity), 0 (even) or 1 (odd).
        :param stop: Number of stop bits. It can be 1 or 2.
        :param tx_pin: Specify the TX PIN to use.
        :param rx_pin: Specify the RX PIN to use.
        :param timeout: Specify the time to wait for the first character in seconds.
        :param timeout_char: Specify the time to wait between characters in seconds.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        code = '''
args = {
    'bits': {bits},
    'parity': {parity},
    'stop': {stop},
}

if {tx_pin}:
    args['tx'] = {tx_pin}
if {rx_pin}:
    args['rx'] = {rx_pin}
if {timeout}:
    args['timeout'] = {timeout}
if {timeout_char}:
    args['timeout_char'] = {timeout_char}
'''.format(bits=bits, parity=parity, stop=stop, tx_pin=tx_pin, rx_pin=rx_pin,
           timeout=timeout, timeout_char=timeout_char)

        self.execute(code, **kwargs)

        code = '''
import machine
uart = machine.UART({id}, {baudrate}, **args)
'''.format(id=id, baudrate=baudrate)

        self.execute(code, **kwargs)

    @action
    def uart_close(self, **kwargs):
        """
        Turn off the UART bus.

        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.uart_open` and
            :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        self.uart_open(**kwargs)
        self.execute('uart.deinit()', **kwargs)

    @action
    def uart_read(self, size: Optional[int] = None, **kwargs) -> str:
        """
        Read from a UART interface.

        :param size: Number of bytes to read (default: read all available characters).
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.uart_open` and
            :meth:`platypush.plugins.esp.EspPlugin.execute`.
        :return: String representation of the read bytes, or base64-encoded representation if the
            data can't be decoded to a string.
        """
        self.uart_open(**kwargs)

        code = '''
args = []
if {size}:
    args.append({size})

uart.read(*args)
'''.format(size=size)

        response = self.execute(code, **kwargs).output
        try:
            return response.decode()
        except UnicodeDecodeError:
            return base64.encodebytes(response).decode()

    @action
    def uart_readline(self, **kwargs) -> str:
        """
        Read a line (any character until newline is found) from a UART interface.

        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.uart_open` and
            :meth:`platypush.plugins.esp.EspPlugin.execute`.
        :return: String representation of the read bytes, or base64-encoded representation if the
            data can't be decoded to a string.
        """
        self.uart_open(**kwargs)
        response = self.execute('uart.readline()', **kwargs).output

        try:
            return response.decode()
        except UnicodeDecodeError:
            return base64.encodebytes(response).decode()

    @action
    def uart_write(self, data: str, binary: bool = False, **kwargs):
        """
        Write data to the UART bus.

        :param data: Data to be written.
        :param binary: By default data will be treated as a string. Set binary to True if it should
            instead be treated as a base64-encoded binary string to be decoded before being sent.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.uart_open` and
            :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        if binary:
            data = base64.decodebytes(data.encode())
        else:
            data = data.encode()

        data = 'b"' + ''.join(['\\x{:02x}'.format(b) for b in data]) + '"'
        self.uart_open(**kwargs)

        code = 'uart.write({data})'.format(data=data)
        self.execute(code, **kwargs)

    @action
    def uart_send_break(self, **kwargs):
        """
        Send a break condition to a UART bus.
        This drives the bus low for a duration longer than required for a normal transmission of a character.

        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.uart_open` and
            :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        self.uart_open(**kwargs)
        code = 'uart.sendbreak()'
        self.execute(code, **kwargs)

    @action
    def get_freq(self, **kwargs) -> int:
        """
        Get the frequency of the device in Hz.

        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        code = '''
import machine
machine.freq()
'''
        return self.execute(code, **kwargs).output

    @action
    def set_freq(self, freq: int, **kwargs):
        """
        Set the frequency of the device.
        :param freq: New frequency in Hz.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        code = '''
import machine
machine.freq({freq})
'''.format(freq=freq)
        self.execute(code, **kwargs)

    @action
    def reset(self, **kwargs):
        """
        Perform a device reset, similar to the user pushing the RESET button.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        code = '''
import machine
machine.reset()
'''
        return self.execute(code, wait_response=False, **kwargs).output

    @action
    def soft_reset(self, **kwargs):
        """
        Performs a soft reset of the interpreter, deleting all Python objects and resetting the Python heap.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        code = '''
import machine
machine.soft_reset()
'''
        return self.execute(code, wait_response=False, **kwargs).output

    @action
    def disable_irq(self, **kwargs):
        """
        Disable interrupt requests.

        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        code = '''
import machine
machine.disable_irq()
'''
        return self.execute(code, **kwargs).output

    @action
    def enable_irq(self, **kwargs):
        """
        Enable interrupt requests.

        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        code = '''
import machine
machine.enable_irq()
'''
        return self.execute(code, **kwargs).output

    @action
    def sleep(self, seconds: float, **kwargs):
        """
        Perform a software sleep (i.e. ``time.sleep()``).

        :param seconds: Sleep seconds.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        code = '''
import time
time.sleep({sec})
'''.format(sec=seconds)

        return self.execute(code, wait_response=False, **kwargs).output

    @action
    def soft_sleep(self, seconds: Optional[float] = None, **kwargs):
        """
        Stops execution in an attempt to enter a low power state.
        A light-sleep has full RAM and state retention. Upon wake execution is resumed from the point where the sleep
        was requested, with all subsystems operational.

        :param seconds: Sleep seconds (default: sleep until there are some PIN/RTC events to process)
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        code = '''
import machine
machine.lightsleep({msec})
'''.format(msec=int(seconds * 1000) if seconds else '')

        return self.execute(code, wait_response=False, **kwargs).output

    @action
    def deep_sleep(self, seconds: Optional[float] = None, **kwargs):
        """
        Stops execution in an attempt to enter a low power state.
        A deepsleep may not retain RAM or any other state of the system (for example peripherals or network interfaces).
        Upon wake execution is resumed from the main script, similar to a hard or power-on reset.

        :param seconds: Sleep seconds (default: sleep until there are some PIN/RTC events to process)
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        code = '''
import machine
machine.deepsleep({msec})
'''.format(msec=int(seconds * 1000) if seconds else '')

        return self.execute(code, wait_response=False, **kwargs).output

    @action
    def unique_id(self, **kwargs) -> str:
        """
        Get the unique ID of the device.
        t will vary from a board/SoC instance to another, if underlying hardware allows. Length varies by hardware
        (so use substring of a full value if you expect a short ID). In some MicroPython ports, ID corresponds to
        the network MAC address..

        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        code = '''
import machine
print(':'.join(['{:02x}'.format(b) for b in machine.unique_id()]))
'''

        return self.execute(code, **kwargs).output

    @action
    def set_password(self, new_password: str, **kwargs):
        """
        Change the WebREPL password for the device.

        :param new_password: New password.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        code = '''
import webrepl
webrepl._webrepl.password({password})
'''.format(password=new_password)

        return self.execute(code, **kwargs).output

    @action
    def wifi_connect(self, essid: str, passphrase: str, **kwargs):
        """
        Connect the device WiFi interface to the specified access point.

        :param essid: WiFi ESSID.
        :param passphrase: WiFi passphrase.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        code = '''
import network
import time

wlan = network.WLAN(network.STA_IF)
wlan.active(True)
wlan.connect('{essid}', '{passphrase}')

while not wlan.isconnected():
    time.sleep(1)
'''.format(essid=essid, passphrase=passphrase)

        self.execute(code, **kwargs)

    def net_enabled_change(self, net: str, enabled: bool, **kwargs):
        code = '''
import network
wlan = network.WLAN({})
wlan.active({})
'''.format(net, str(enabled))
        self.execute(code, **kwargs)

    def net_ifconfig(self, net: str, ip: Optional[str] = None, netmask: Optional[str] = None,
                     gateway: Optional[str] = None, dns: Optional[str] = None, **kwargs):
        code = '''
import json
import network

wlan = network.WLAN({})
print(json.dumps(list(wlan.ifconfig())))
'''.format(net)

        config = self.execute(code, **kwargs).output
        if ip:
            config[0] = ip
        if netmask:
            config[1] = netmask
        if gateway:
            config[2] = gateway
        if dns:
            config[3] = dns

        code = 'wlan.ifconfig({})'.format(tuple(config))
        self.execute(code, **kwargs)

    @action
    def wifi_config(self, ip: Optional[str] = None, netmask: Optional[str] = None,
                    gateway: Optional[str] = None, dns: Optional[str] = None, **kwargs) \
            -> Optional[EspWifiConfigResult]:
        """
        Get or set network properties for the WiFi interface.
        If called with no arguments it will return the configuration of the interface.

        :param ip: IP address.
        :param netmask: Netmask.
        :param gateway: Default gateway address.
        :param dns: Default DNS address.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        if ip or netmask or gateway or dns:
            self.net_ifconfig(net_type='network.STA_IF', ip=ip, netmask=netmask, gateway=gateway, dns=dns, **kwargs)
            return

        return self.net_config('network.STA_IF', **kwargs)

    @action
    def ap_config(self, ip: Optional[str] = None, netmask: Optional[str] = None,
                  gateway: Optional[str] = None, dns: Optional[str] = None, essid: Optional[str] = None,
                  passphrase: Optional[str] = None, channel: Optional[int] = None,
                  hidden: Optional[bool] = None, **kwargs) -> Optional[EspWifiConfigResult]:
        """
        Get or set network properties for the WiFi access point interface.
        If called with no arguments it will return the configuration of the interface.

        :param ip: IP address.
        :param netmask: Netmask.
        :param gateway: Default gateway address.
        :param dns: Default DNS address.
        :param essid: ESSID of the access point.
        :param passphrase: Password/passphrase.
        :param channel: WiFi channel.
        :param hidden: Whether the network is hidden.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        has_args = False
        self.execute(code='import network; ap = network.WLAN(network.AP_IF)'.format(essid=essid), **kwargs)

        if ip or netmask or gateway or dns:
            self.net_ifconfig(net_type='network.AP_IF', ip=ip, netmask=netmask, gateway=gateway, dns=dns, **kwargs)
            has_args = True
        if essid:
            self.execute(code='ap.config(essid="{essid}")'.format(essid=essid), **kwargs)
            has_args = True
        if passphrase:
            self.execute(code='ap.config(password="{passphrase}")'.format(passphrase=passphrase), **kwargs)
            has_args = True
        if channel:
            self.execute(code='ap.config(channel={channel})'.format(channel=channel), **kwargs)
            has_args = True
        if hidden is not None:
            self.execute(code='ap.config(hidden={hidden})'.format(hidden=str(hidden)), **kwargs)
            has_args = True

        if has_args:
            return

        return self.net_config('network.AP_IF', **kwargs)

    @action
    def wifi_enable(self, **kwargs):
        """
        Enable the device WiFi interface.

        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        self.net_enabled_change('network.STA_IF', True, **kwargs)

    @action
    def wifi_disable(self, **kwargs):
        """
        Disable the device WiFi interface.

        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        self.net_enabled_change('network.STA_IF', False, **kwargs)

    @action
    def wifi_disconnect(self, **kwargs):
        """
        Disconnect from the currently connected WiFi network

        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        code = '''
import network
wlan = network.WLAN(network.STA_IF)
wlan.disconnect()
'''
        self.execute(code, **kwargs)

    @action
    def ap_enable(self, **kwargs):
        """
        Enable the device WiFi access point interface.

        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        self.net_enabled_change('network.AP_IF', True, **kwargs)

    @action
    def ap_disable(self, **kwargs):
        """
        Disable the device WiFi access point interface.

        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        self.net_enabled_change('network.AP_IF', False, **kwargs)

    def net_config(self, net_type: str, **kwargs) -> Optional[EspWifiConfigResult]:
        # Split the code into multiple execution to overcome the size limitation of the input ESP WebREPL buffer.
        code = '''
import network
import json

net = network.WLAN(''' + net_type + ''')
ifconfig = net.ifconfig()
mac = ':'.join([hex(c)[2:] for c in net.config('mac')])'''
        self.execute(code, **kwargs)

        code = '''
active = net.active()
essid = net.config('essid')
channel = net.config('channel')
try:
    hidden = net.config('hidden')
except:
    hidden = None

connected = net.isconnected()'''
        self.execute(code, **kwargs)

        code = '''
config = {
    'ip': ifconfig[0],
    'netmask': ifconfig[1],
    'gateway': ifconfig[2],
    'dns': ifconfig[3],
    'mac': mac,
    'connected': connected,
    'active': active,
    'essid': essid,
    'channel': channel,
    'hidden': hidden,
}

print(json.dumps(config))
'''

        net = self.execute(code, **kwargs).output
        if not net:
            return
        return EspWifiConfigResult(**net)

    @action
    def wifi_scan(self, **kwargs) -> List[EspWifiScanResult]:
        """
        Scan the available networks.

        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        code = '''
import network
import json

wlan = network.WLAN(network.STA_IF)
wlan.active(True)
print(json.dumps([{
    'essid':net[0].decode(),
    'bssid':''.join(['\\\\x' + hex(c)[2:] for c in net[1]]),
    'channel': net[2],
    'rssi': net[3],
    'auth_mode': net[4],
    'hidden': bool(net[5])
} for net in wlan.scan()]))
'''
        results = self.execute(code, **kwargs).output
        if not results:
            return []

        return [EspWifiScanResult(**network) for network in results]

    def open_db(self, dbfile: str, **kwargs):
        code = '''
import btree

try:
    dbfile = open('{dbfile}', 'r+b')
except OSError:
    dbfile = open('{dbfile}', 'w+b')

db = btree.open(dbfile)
'''.format(dbfile=dbfile)

        self.close_db(dbfile, **kwargs)
        self.execute(code, **kwargs)

    def close_db(self, dbfile: str, **kwargs):
        code = '''
try:
    db.close()
    dbfile.close()
    db = None
    dbfile = None
except:
    pass
'''.format(dbfile=dbfile)

        self.execute(code, **kwargs)

    @staticmethod
    def string_quote(s: str):
        return s.replace("'", "\\'")

    @action
    def db_set(self, dbfile: str, key: str, value: Any, **kwargs):
        """
        Set a value on an internal B-Tree file database.

        :param dbfile: Database file name.
        :param key: Key to set.
        :param value: Value to set.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        code = '''
db[b'{key}'] = b'{value}'
db.flush()
'''.format(key=self.string_quote(key), value=self.string_quote(str(value)))

        try:
            self.open_db(dbfile, **kwargs)
            self.execute(code, **kwargs)
        finally:
            self.close_db(dbfile, **kwargs)

    @action
    def db_get(self, dbfile: str, key: str, **kwargs) -> Any:
        """
        Set a value on an internal B-Tree file database.

        :param dbfile: Database file name.
        :param key: Key to set.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        :return: Whichever value is stored as output, or null if the key doesn't exist.
        """
        code = '''
try:
    print(db[b'{key}'].decode())
except KeyError:
    pass
'''.format(key=self.string_quote(key))

        try:
            self.open_db(dbfile, **kwargs)
            response = self.execute(code, **kwargs)
            return response
        finally:
            self.close_db(dbfile, **kwargs)

    @action
    def db_keys(self, dbfile: str, **kwargs) -> List[str]:
        """
        Get the list of keys stored on an internal B-Tree file database.

        :param dbfile: Database file name.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        code = '''
import json
print(json.dumps([k.decode() for k in db.keys()]))
'''

        try:
            self.open_db(dbfile, **kwargs)
            return self.execute(code, **kwargs).output
        finally:
            self.close_db(dbfile, **kwargs)

    @action
    def db_values(self, dbfile: str, **kwargs) -> List[str]:
        """
        Get the list of item values stored on an internal B-Tree file database.

        :param dbfile: Database file name.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        code = '''
import json
print(json.dumps([k.decode() for k in db.values()]))
'''

        try:
            self.open_db(dbfile, **kwargs)
            return self.execute(code, **kwargs).output
        finally:
            self.close_db(dbfile, **kwargs)

    @action
    def db_items(self, dbfile: str, **kwargs) -> Dict[str, str]:
        """
        Get a key->value mapping of the items stored in a B-Tree file database.

        :param dbfile: Database file name.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        :return: Whichever value is stored as output, or null if the key doesn't exist.
        """
        code = '''
import json
print(json.dumps({k.decode(): v.decode() for k, v in db.items()}))
'''

        try:
            self.open_db(dbfile, **kwargs)
            return self.execute(code, **kwargs).output
        finally:
            self.close_db(dbfile, **kwargs)

    @action
    def set_ntp_time(self, **kwargs):
        """
        Set the device time using an NTP server.

        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        code = '''
import ntptime
ntptime.settime()
'''

        self.execute(code, **kwargs)

    @action
    def listdir(self, directory: str = '/', **kwargs) -> List[str]:
        """
        List the content of a directory.

        :param directory: Directory name (default: root).
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        directory = self.string_quote(directory)
        code = '''
import os
import json

print(json.dumps(os.listdir('{dir}')))
'''.format(dir=directory)

        return self.execute(code, **kwargs).output

    @action
    def chdir(self, directory: str, **kwargs):
        """
        Move to the specified directory.

        :param directory: Directory name.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        directory = self.string_quote(directory)
        code = '''
import os
os.chdir('{dir}')
'''.format(dir=directory)

        self.execute(code, **kwargs)

    @action
    def mkdir(self, directory: str, **kwargs):
        """
        Create a directory.

        :param directory: Directory name.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        directory = self.string_quote(directory)
        code = '''
import os
os.mkdir('{dir}')
'''.format(dir=directory)

        self.execute(code, **kwargs)

    @action
    def rmdir(self, directory: str, **kwargs):
        """
        Remove a directory.

        :param directory: Directory name.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        directory = self.string_quote(directory)
        code = '''
import os
os.rmdir('{dir}')
'''.format(dir=directory)

        self.execute(code, **kwargs)

    @action
    def rename(self, name: str, new_name: str, **kwargs):
        """
        Rename a file or directory.

        :param name: Current resource name.
        :param new_name: New resource name.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        name = self.string_quote(name)
        new_name = self.string_quote(new_name)
        code = '''
import os
os.rename('{old}', '{new}')
'''.format(old=name, new=new_name)

        self.execute(code, **kwargs)

    @action
    def remove(self, file: str, **kwargs):
        """
        Remove a file.

        :param file: File name/path.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        file = self.string_quote(file)
        code = '''
import os
os.remove('{file}')
'''.format(file=file)

        self.execute(code, **kwargs)

    @action
    def urandom(self, size: int = 1, **kwargs) -> List[int]:
        """
        Get randomly generated bytes.

        :param size: Number of random bytes to be generated (default: 1).
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        code = '''
import os
import json

print([b for b in os.urandom({size})])
'''.format(size=size)

        return self.execute(code, **kwargs).output

    @action
    def file_get(self, file: str, binary: bool = False, timeout: Optional[float] = 60.0, **kwargs) -> str:
        """
        Get the content of a file on the board.

        :param file: File name/path to get from the device.
        :param binary: If True, then the base64-encoded content of the file will be returned.
        :param timeout: File transfer timeout (default: one minute).
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.connect`.
        """
        device = self._get_device(**kwargs)
        host = device['host']
        port = device['port']
        self.connect(host=host, port=port, password=device['password'])
        conn = self._get_connection(host=host, port=port)

        with io.BytesIO() as buffer:
            conn.file_download(file, buffer, timeout=timeout)
            data = buffer.getvalue()

        if binary:
            data = base64.encodebytes(data).decode()
        else:
            data = data.decode()

        return data

    @action
    def file_upload(self, source: str, destination: Optional[str] = None, timeout: Optional[float] = 60.0, **kwargs):
        """
        Upload a file to the board.

        :param source: Path of the local file to copy.
        :param destination: Target file name (default: a filename will be created under the board's
            root folder with the same name as the source file).
        :param timeout: File transfer timeout (default: one minute).
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.connect`.
        """
        device = self._get_device(**kwargs)
        host = device['host']
        port = device['port']
        self.connect(host=host, port=port, password=device['password'])
        conn = self._get_connection(host=host, port=port)
        conn.file_upload(source=source, destination=destination, timeout=timeout)

    @action
    def file_download(self, source: str, destination: str, timeout: Optional[float] = 60.0, **kwargs):
        """
        Download a file from the board to the local machine.

        :param source: Name or path of the file to download from the device.
        :param destination: Target directory or path on the local machine.
        :param timeout: File transfer timeout (default: one minute).
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        destination = os.path.abspath(os.path.expanduser(destination))
        if os.path.isdir(destination):
            filename = os.path.basename(source)
            destination = os.path.join(destination, filename)

        device = self._get_device(**kwargs)
        host = device['host']
        port = device['port']
        self.connect(host=host, port=port, password=device['password'])
        conn = self._get_connection(host=host, port=port)

        with open(destination, 'wb') as f:
            conn.file_download(source, f, timeout=timeout)

    def _dht_get_value(self, pin: Union[int, str], dht_type: int, value: str, **kwargs) -> float:
        device = self._get_device(**kwargs)
        pin = device.get_pin(pin)
        code = '''
import machine
import dht

dht_sensor = dht.DHT{type}(machine.Pin({pin}))
dht_sensor.measure()
dht_sensor.{value}()
'''.format(pin=pin, type=dht_type, value=value)

        return self.execute(code, **kwargs).output

    @action
    def dht11_get_temperature(self, pin: Union[int, str], **kwargs) -> float:
        """
        Get the temperature value in Celsius from a connected DHT11 sensor.

        :param pin: GPIO PIN number or configured name where the sensor is connected.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        return self._dht_get_value(pin=pin, dht_type=11, value='temperature', **kwargs)

    @action
    def dht11_get_humidity(self, pin: Union[int, str], **kwargs) -> float:
        """
        Get the humidity value in percentage (0-100) from a connected DHT11 sensor.

        :param pin: GPIO PIN number or configured name where the sensor is connected.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        return self._dht_get_value(pin=pin, dht_type=11, value='humidity', **kwargs)

    @action
    def dht22_get_temperature(self, pin: Union[int, str], **kwargs) -> float:
        """
        Get the temperature value in Celsius from a connected DHT22 sensor.

        :param pin: GPIO PIN number or configured name where the sensor is connected.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        return self._dht_get_value(pin=pin, dht_type=22, value='temperature', **kwargs)

    @action
    def dht22_get_humidity(self, pin: Union[int, str], **kwargs) -> float:
        """
        Get the humidity value in percentage (0-100) from a connected DHT22 sensor.

        :param pin: GPIO PIN number or configured name where the sensor is connected.
        :param kwargs: Parameters to pass to :meth:`platypush.plugins.esp.EspPlugin.execute`.
        """
        return self._dht_get_value(pin=pin, dht_type=22, value='humidity', **kwargs)


# vim:sw=4:ts=4:et:



from platypush.message.event import Event


class CameraEvent(Event):
    """ Base class for camera events """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class CameraRecordingStartedEvent(CameraEvent):
    """
    Event triggered when a new recording starts
    """

    def __init__(self, device_id, filename=None, *args, **kwargs):
        super().__init__(*args, device_id=device_id, filename=filename, **kwargs)


class CameraRecordingStoppedEvent(CameraEvent):
    """
    Event triggered when a recording stops
    """

    def __init__(self, device_id, *args, **kwargs):
        super().__init__(*args, device_id=device_id, **kwargs)


class CameraVideoRenderedEvent(CameraEvent):
    """
    Event triggered when a sequence of frames has been rendered into a video
    """

    def __init__(self, filename=None, *args, **kwargs):
        super().__init__(*args, filename=filename, **kwargs)


class CameraPictureTakenEvent(CameraEvent):
    """
    Event triggered when a snapshot has been taken
    """

    def __init__(self, filename=None, *args, **kwargs):
        super().__init__(*args, filename=filename, **kwargs)


class CameraFrameCapturedEvent(CameraEvent):
    """
    Event triggered when a camera frame has been captured
    """

    disable_logging = True

    def __init__(self, filename=None, *args, **kwargs):
        super().__init__(*args, filename=filename,
                         disable_logging=kwargs.pop('disable_logging')
                         if 'disable_logging' in kwargs else self.disable_logging,
                         **kwargs)


# vim:sw=4:ts=4:et:



"""
Platypush

.. moduleauthor:: Fabio Manganiello <blacklight86@gmail.com>
.. license: MIT
"""

import argparse
import logging
import os
import sys

from .bus.redis import RedisBus
from .config import Config
from .context import register_backends
from .cron.scheduler import CronScheduler
from .event.processor import EventProcessor
from .logger import Logger
from .message.event import Event, StopEvent
from .message.event.application import ApplicationStartedEvent, ApplicationStoppedEvent
from .message.request import Request
from .message.response import Response
from .utils import set_thread_name


__author__ = 'Fabio Manganiello <blacklight86@gmail.com>'
__version__ = '0.12.3'

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)


class Daemon:
    """ Main class for the Platypush daemon """

    # Configuration file (default: either ~/.config/platypush/config.yaml or
    # /etc/platypush/config.yaml
    config_file = None

    # Application bus. It's an internal queue where:
    # - backends will post the messages they receive
    # - plugins will post the responses they process
    bus = None

    pidfile = None

    # backend_name => backend_obj map
    backends = None

    # number of executions retries before a request fails
    n_tries = 2

    def __init__(self, config_file=None, pidfile=None, requests_to_process=None,
                 no_capture_stdout=False, no_capture_stderr=False):
        """
        Constructor
        Params:
            config_file -- Configuration file override (default: None)
            pidfile -- File where platypush will store its PID upon launch,
                       useful if you're planning to integrate the application
                       within a service or a launcher script (default: None)
            requests_to_process -- Exit after processing the specified number
                                   of requests (default: None, loop forever)
            no_capture_stdout -- Set to true if you want to disable the stdout
                                 capture by the logging system
            no_capture_stderr -- Set to true if you want to disable the stderr
                                 capture by the logging system
        """

        if pidfile:
            self.pidfile = pidfile
            with open(self.pidfile, 'w') as f:
                f.write(str(os.getpid()))

        self.config_file = config_file
        Config.init(self.config_file)
        logging.basicConfig(**Config.get('logging'))

        self.no_capture_stdout = no_capture_stdout
        self.no_capture_stderr = no_capture_stderr
        self.event_processor = EventProcessor()
        self.requests_to_process = requests_to_process
        self.processed_requests = 0

    @classmethod
    def build_from_cmdline(cls, args):
        """
        Build the app from command line arguments.
        Params:
            args -- Your sys.argv[1:] [List of strings]
        """
        parser = argparse.ArgumentParser()
        parser.add_argument('--config', '-c', dest='config', required=False,
                            default=None, help=cls.config_file.__doc__)
        parser.add_argument('--pidfile', '-P', dest='pidfile', required=False,
                            default=None, help="File where platypush will " +
                            "store its PID, useful if you're planning to " +
                            "integrate it in a service")
        parser.add_argument('--no-capture-stdout', dest='no_capture_stdout',
                            required=False, action='store_true',
                            help="Set this flag if you have max stack depth " +
                            "exceeded errors so stdout won't be captured by " +
                            "the logging system")
        parser.add_argument('--no-capture-stderr', dest='no_capture_stderr',
                            required=False, action='store_true',
                            help="Set this flag if you have max stack depth " +
                            "exceeded errors so stderr won't be captured by " +
                            "the logging system")

        opts, args = parser.parse_known_args(args)
        return cls(config_file=opts.config, pidfile=opts.pidfile,
                   no_capture_stdout=opts.no_capture_stdout,
                   no_capture_stderr=opts.no_capture_stderr)

    def on_message(self):
        """
        Default message handler
        """

        def _f(msg):
            """
            on_message closure
            Params:
                msg -- platypush.message.Message instance
            """

            if isinstance(msg, Request):
                try:
                    msg.execute(n_tries=self.n_tries)
                except PermissionError:
                    LOGGER.info('Dropped unauthorized request: {}'.format(msg))

                self.processed_requests += 1
                if self.requests_to_process \
                        and self.processed_requests >= self.requests_to_process:
                    self.stop_app()
            elif isinstance(msg, Response):
                LOGGER.info('Received response: {}'.format(msg))
            elif isinstance(msg, StopEvent) and msg.targets_me():
                LOGGER.info('Received STOP event: {}'.format(msg))
                self.stop_app()
            elif isinstance(msg, Event):
                if not msg.disable_logging:
                    LOGGER.info('Received event: {}'.format(msg))
                self.event_processor.process_event(msg)

        return _f

    def stop_app(self):
        """ Stops the backends and the bus """
        for backend in self.backends.values():
            backend.stop()
        self.bus.stop()

    def start(self):
        """ Start the daemon """
        if not self.no_capture_stdout:
            sys.stdout = Logger(LOGGER.info)
        if not self.no_capture_stderr:
            sys.stderr = Logger(LOGGER.warning)

        set_thread_name('platypush')

        print('---- Starting platypush v.{}'.format(__version__))

        redis_conf = Config.get('backend.redis') or {}
        self.bus = RedisBus(on_message=self.on_message(),
                            **redis_conf.get('redis_args', {}))

        # Initialize the backends and link them to the bus
        self.backends = register_backends(bus=self.bus, global_scope=True)

        # Start the backend threads
        for backend in self.backends.values():
            backend.start()

        # Start the cron scheduler
        if Config.get_cronjobs():
            CronScheduler(jobs=Config.get_cronjobs()).start()

        self.bus.post(ApplicationStartedEvent())

        # Poll for messages on the bus
        try:
            self.bus.poll()
        except KeyboardInterrupt:
            LOGGER.info('SIGINT received, terminating application')
        finally:
            self.bus.post(ApplicationStoppedEvent())
            self.stop_app()


def main():
    """
    Platypush daemon main
    """

    app = Daemon.build_from_cmdline(sys.argv[1:])
    app.start()


# vim:sw=4:ts=4:et:



import datetime
import re
import time

from platypush.context import get_plugin, get_bus
from platypush.plugins import action
from platypush.plugins.media import MediaPlugin
from platypush.utils import get_mime_type
from platypush.message.event.media import MediaPlayEvent, MediaPlayRequestEvent, \
    MediaStopEvent, MediaPauseEvent, NewPlayingMediaEvent, MediaVolumeChangedEvent, MediaSeekEvent


def convert_status(status):
    attrs = [a for a in dir(status) if not a.startswith('_')
             and not callable(getattr(status, a))]

    renamed_attrs = {
        'current_time': 'position',
        'player_state': 'state',
        'supports_seek': 'seekable',
        'volume_level': 'volume',
        'volume_muted': 'mute',
        'content_id': 'url',
    }

    ret = {}
    for attr in attrs:
        value = getattr(status, attr)
        if attr == 'volume_level':
            value *= 100
        if attr == 'player_state':
            value = value.lower()
            if value == 'paused':
                value = 'pause'
            if value == 'playing':
                value = 'play'
        if isinstance(value, datetime.datetime):
            value = value.isoformat()

        if attr in renamed_attrs:
            ret[renamed_attrs[attr]] = value
        else:
            ret[attr] = value

    return ret


def post_event(evt_type, **evt):
    bus = get_bus()
    bus.post(evt_type(player=evt.get('device'), plugin='media.chromecast', **evt))


class MediaChromecastPlugin(MediaPlugin):
    """
    Plugin to interact with Chromecast devices

    Supported formats:

        * HTTP media URLs
        * YouTube URLs
        * Plex (through ``media.plex`` plugin, experimental)

    Requires:

        * **pychromecast** (``pip install pychromecast``)
    """

    STREAM_TYPE_UNKNOWN = "UNKNOWN"
    STREAM_TYPE_BUFFERED = "BUFFERED"
    STREAM_TYPE_LIVE = "LIVE"

    class MediaListener:
        def __init__(self, name, cast):
            self.name = name
            self.cast = cast
            self.status = convert_status(cast.media_controller.status)
            self.last_status_timestamp = time.time()

        def new_media_status(self, status):
            status = convert_status(status)

            if status.get('url') and status.get('url') != self.status.get('url'):
                post_event(NewPlayingMediaEvent, resource=status['url'],
                           title=status.get('title'), device=self.name)
            if status.get('state') != self.status.get('state'):
                if status.get('state') == 'play':
                    post_event(MediaPlayEvent, resource=status['url'], device=self.name)
                elif status.get('state') == 'pause':
                    post_event(MediaPauseEvent, resource=status['url'], device=self.name)
                elif status.get('state') in ['stop', 'idle']:
                    post_event(MediaStopEvent, device=self.name)
            if status.get('volume') != self.status.get('volume'):
                post_event(MediaVolumeChangedEvent, volume=status.get('volume'), device=self.name)
            if abs(status.get('position') - self.status.get('position')) > time.time() - self.last_status_timestamp + 5:
                post_event(MediaSeekEvent, position=status.get('position'), device=self.name)

            self.last_status_timestamp = time.time()
            self.status = status

    class SubtitlesAsyncHandler:
        def __init__(self, mc, subtitle_id):
            self.mc = mc
            self.subtitle_id = subtitle_id
            self.initialized = False

        # pylint: disable=unused-argument
        def new_media_status(self, status):
            if self.subtitle_id and not self.initialized:
                self.mc.update_status()
                self.mc.enable_subtitle(self.subtitle_id)
                self.initialized = True

    def __init__(self, chromecast=None, *args, **kwargs):
        """
        :param chromecast: Default Chromecast to cast to if no name is specified
        :type chromecast: str
        """

        super().__init__(*args, **kwargs)

        self._is_local = False
        self.chromecast = chromecast
        self.chromecasts = {}
        self._media_listeners = {}

    @action
    def get_chromecasts(self, tries=2, retry_wait=10, timeout=60,
                        blocking=True, callback=None):
        """
        Get the list of Chromecast devices

        :param tries: Number of retries (default: 2)
        :type tries: int

        :param retry_wait: Number of seconds between retries (default: 10 seconds)
        :type retry_wait: int

        :param timeout: Timeout before failing the call (default: 60 seconds)
        :type timeout: int

        :param blocking: If true, then the function will block until all the Chromecast
            devices have been scanned. If false, then the provided callback function will be
            invoked when a new device is discovered
        :type blocking: bool

        :param callback: If blocking is false, then you can provide a callback function that
            will be invoked when a new device is discovered
        :type callback: func
        """

        import pychromecast
        self.chromecasts.update({
            cast.device.friendly_name: cast
            for cast in pychromecast.get_chromecasts(tries=tries, retry_wait=retry_wait,
                                                     timeout=timeout, blocking=blocking,
                                                     callback=callback)
        })

        for name, cast in self.chromecasts.items():
            self._update_listeners(name, cast)

        return [{
            'type': cc.cast_type,
            'name': cc.name,
            'manufacturer': cc.device.manufacturer,
            'model_name': cc.model_name,
            'uuid': str(cc.device.uuid),
            'address': cc.host,
            'port': cc.port,

            'status': ({
                'app': {
                    'id': cc.app_id,
                    'name': cc.app_display_name,
                },

                'media': self.status(cc.name).output,
                'is_active_input': cc.status.is_active_input,
                'is_stand_by': cc.status.is_stand_by,
                'is_idle': cc.is_idle,
                'namespaces': cc.status.namespaces,
                'volume': round(100*cc.status.volume_level, 2),
                'muted': cc.status.volume_muted,
            } if cc.status else {}),
        } for cc in self.chromecasts.values()]

    def _update_listeners(self, name, cast):
        if name not in self._media_listeners:
            cast.start()
            self._media_listeners[name] = self.MediaListener(name=name, cast=cast)
            cast.media_controller.register_status_listener(self._media_listeners[name])

    def get_chromecast(self, chromecast=None, n_tries=2):
        import pychromecast
        if isinstance(chromecast, pychromecast.Chromecast):
            return chromecast

        if not chromecast:
            if not self.chromecast:
                raise RuntimeError('No Chromecast specified nor default Chromecast configured')
            chromecast = self.chromecast

        if chromecast not in self.chromecasts:
            casts = {}
            while n_tries > 0:
                n_tries -= 1
                casts.update({
                    cast.device.friendly_name: cast
                    for cast in pychromecast.get_chromecasts()
                })

                if chromecast in casts:
                    self.chromecasts.update(casts)
                    break

            if chromecast not in self.chromecasts:
                raise RuntimeError('Device {} not found'.format(chromecast))

        cast = self.chromecasts[chromecast]
        cast.wait()
        return cast

    @action
    def play(self, resource, content_type=None, chromecast=None, title=None,
             image_url=None, autoplay=True, current_time=0,
             stream_type=STREAM_TYPE_BUFFERED, subtitles=None,
             subtitles_lang='en-US', subtitles_mime='text/vtt',
             subtitle_id=1):
        """
        Cast media to a visible Chromecast

        :param resource: Media to cast
        :type resource: str

        :param content_type: Content type as a MIME type string
        :type content_type: str

        :param chromecast: Chromecast to cast to. If none is specified, then the default configured Chromecast
            will be used.
        :type chromecast: str

        :param title: Optional title
        :type title: str

        :param image_url: URL of the image to use for the thumbnail
        :type image_url: str

        :param autoplay: Set it to false if you don't want the content to start playing immediately (default: true)
        :type autoplay: bool

        :param current_time: Time to start the playback in seconds (default: 0)
        :type current_time: int

        :param stream_type: Type of stream to cast. Can be BUFFERED (default), LIVE or UNKNOWN
        :type stream_type: str

        :param subtitles: URL of the subtitles to be shown
        :type subtitles: str

        :param subtitles_lang: Subtitles language (default: en-US)
        :type subtitles_lang: str

        :param subtitles_mime: Subtitles MIME type (default: text/vtt)
        :type subtitles_mime: str

        :param subtitle_id: ID of the subtitles to be loaded (default: 1)
        :type subtitle_id: int
        """

        from pychromecast.controllers.youtube import YouTubeController
        if not chromecast:
            chromecast = self.chromecast

        post_event(MediaPlayRequestEvent, resource=resource, device=chromecast)
        cast = self.get_chromecast(chromecast)

        mc = cast.media_controller
        yt = self._get_youtube_url(resource)

        if yt:
            self.logger.info('Playing YouTube video {} on {}'.format(
                yt, chromecast))

            hndl = YouTubeController()
            cast.register_handler(hndl)
            hndl.update_screen_id()
            return hndl.play_video(yt)

        resource = self._get_resource(resource)

        if resource.startswith('magnet:?'):
            player_args = {'chromecast': chromecast}
            return get_plugin('media.webtorrent').play(resource,
                                                       player='chromecast',
                                                       **player_args)

        if not content_type:
            content_type = get_mime_type(resource)

        if not content_type:
            raise RuntimeError('content_type required to process media {}'.
                               format(resource))

        if not resource.startswith('http://') and \
                not resource.startswith('https://'):
            resource = self.start_streaming(resource).output['url']
            self.logger.info('HTTP media stream started on {}'.format(resource))

        self.logger.info('Playing {} on {}'.format(resource, chromecast))

        mc.play_media(resource, content_type, title=title, thumb=image_url,
                      current_time=current_time, autoplay=autoplay,
                      stream_type=stream_type, subtitles=subtitles,
                      subtitles_lang=subtitles_lang,
                      subtitles_mime=subtitles_mime, subtitle_id=subtitle_id)

        if subtitles:
            mc.register_status_listener(self.SubtitlesAsyncHandler(mc, subtitle_id))

        mc.block_until_active()

        if self.volume:
            self.set_volume(volume=self.volume, chromecast=chromecast)

        return self.status(chromecast=chromecast)

    @classmethod
    def _get_youtube_url(cls, url):
        m = re.match('https?://www.youtube.com/watch\?v=([^&]+).*', url)
        if m:
            return m.group(1)

        m = re.match('youtube:video:(.*)', url)
        if m:
            return m.group(1)

        return None

    @action
    def load(self, *args, **kwargs):
        return self.play(*args, **kwargs)

    @action
    def pause(self, chromecast=None):
        chromecast = chromecast or self.chromecast
        cast = self.get_chromecast(chromecast)

        if cast.media_controller.is_paused:
            cast.media_controller.play()
        elif cast.media_controller.is_playing:
            cast.media_controller.pause()

        cast.wait()
        return self.status(chromecast=chromecast)

    @action
    def stop(self, chromecast=None):
        chromecast = chromecast or self.chromecast
        cast = self.get_chromecast(chromecast)
        cast.media_controller.stop()
        cast.wait()

        return self.status(chromecast=chromecast)

    @action
    def rewind(self, chromecast=None):
        chromecast = chromecast or self.chromecast
        cast = self.get_chromecast(chromecast)
        cast.media_controller.rewind()
        cast.wait()
        return self.status(chromecast=chromecast)

    @action
    def set_position(self, position, chromecast=None):
        cast = self.get_chromecast(chromecast or self.chromecast)
        cast.media_controller.seek(position)
        cast.wait()
        return self.status(chromecast=chromecast)

    @action
    def seek(self, position, chromecast=None):
        return self.forward(chromecast=chromecast, offset=position)

    @action
    def back(self, chromecast=None, offset=60):
        cast = self.get_chromecast(chromecast or self.chromecast)
        mc = cast.media_controller
        if mc.status.current_time:
            mc.seek(mc.status.current_time-offset)
            cast.wait()

        return self.status(chromecast=chromecast)

    @action
    def forward(self, chromecast=None, offset=60):
        cast = self.get_chromecast(chromecast or self.chromecast)
        mc = cast.media_controller
        if mc.status.current_time:
            mc.seek(mc.status.current_time+offset)
            cast.wait()

        return self.status(chromecast=chromecast)

    @action
    def is_playing(self, chromecast=None):
        return self.get_chromecast(chromecast or self.chromecast).media_controller.is_playing

    @action
    def is_paused(self, chromecast=None):
        return self.get_chromecast(chromecast or self.chromecast).media_controller.is_paused

    @action
    def is_idle(self, chromecast=None):
        return self.get_chromecast(chromecast or self.chromecast).media_controller.is_idle

    @action
    def list_subtitles(self, chromecast=None):
        return self.get_chromecast(chromecast or self.chromecast) \
            .media_controller.subtitle_tracks

    @action
    def enable_subtitles(self, chromecast=None, track_id=None):
        mc = self.get_chromecast(chromecast or self.chromecast).media_controller
        if track_id is not None:
            return mc.enable_subtitle(track_id)
        elif mc.subtitle_tracks:
            return mc.enable_subtitle(mc.subtitle_tracks[0].get('trackId'))

    @action
    def disable_subtitles(self, chromecast=None, track_id=None):
        mc = self.get_chromecast(chromecast or self.chromecast).media_controller
        if track_id:
            return mc.disable_subtitle(track_id)
        elif mc.current_subtitle_tracks:
            return mc.disable_subtitle(mc.current_subtitle_tracks[0])

    @action
    def toggle_subtitles(self, chromecast=None):
        mc = self.get_chromecast(chromecast or self.chromecast).media_controller
        all_subs = mc.status.subtitle_tracks
        cur_subs = mc.status.status.current_subtitle_tracks

        if cur_subs:
            return self.disable_subtitles(chromecast, cur_subs[0])
        else:
            return self.enable_subtitles(chromecast, all_subs[0].get('trackId'))

    @action
    def status(self, chromecast=None):
        cast = self.get_chromecast(chromecast or self.chromecast)
        status = cast.media_controller.status
        return convert_status(status)

    @action
    def disconnect(self, chromecast=None, timeout=None, blocking=True):
        """
        Disconnect a Chromecast and wait for it to terminate

        :param chromecast: Chromecast to cast to. If none is specified, then the default configured Chromecast
            will be used.
        :type chromecast: str

        :param timeout: Number of seconds to wait for disconnection (default: None: block until termination)
        :type timeout: float

        :param blocking: If set (default), then the code will wait until disconnection, otherwise it will return
            immediately.
        :type blocking: bool
        """

        cast = self.get_chromecast(chromecast)
        cast.disconnect(timeout=timeout, blocking=blocking)

    @action
    def join(self, chromecast=None, timeout=None):
        """
        Blocks the thread until the Chromecast connection is terminated.

        :param chromecast: Chromecast to cast to. If none is specified, then the default configured Chromecast
            will be used.
        :type chromecast: str

        :param timeout: Number of seconds to wait for disconnection (default: None: block until termination)
        :type timeout: float
        """

        cast = self.get_chromecast(chromecast)
        cast.join(timeout=timeout)

    @action
    def quit(self, chromecast=None):
        """
        Exits the current app on the Chromecast

        :param chromecast: Chromecast to cast to. If none is specified, then the default configured Chromecast
            will be used.
        :type chromecast: str
        """

        cast = self.get_chromecast(chromecast)
        cast.quit_app()

    @action
    def reboot(self, chromecast=None):
        """
        Reboots the Chromecast

        :param chromecast: Chromecast to cast to. If none is specified, then the default configured Chromecast
            will be used.
        :type chromecast: str
        """

        cast = self.get_chromecast(chromecast)
        cast.reboot()

    @action
    def set_volume(self, volume, chromecast=None):
        """
        Set the Chromecast volume

        :param volume: Volume to be set, between 0 and 100
        :type volume: float

        :param chromecast: Chromecast to cast to. If none is specified, then the default configured Chromecast
            will be used.
        :type chromecast: str
        """

        chromecast = chromecast or self.chromecast
        cast = self.get_chromecast(chromecast)
        cast.set_volume(volume/100)
        cast.wait()
        status = self.status(chromecast=chromecast)
        status.output['volume'] = volume
        return status

    @action
    def volup(self, chromecast=None, step=10):
        """
        Turn up the Chromecast volume by 10% or step.

        :param chromecast: Chromecast to cast to. If none is specified, then the default configured Chromecast
            will be used.
        :type chromecast: str

        :param step: Volume increment between 0 and 100 (default: 100%)
        :type step: float
        """

        chromecast = chromecast or self.chromecast
        cast = self.get_chromecast(chromecast)
        step /= 100
        cast.volume_up(min(step, 1))
        cast.wait()
        return self.status(chromecast=chromecast)

    @action
    def voldown(self, chromecast=None, step=10):
        """
        Turn down the Chromecast volume by 10% or step.

        :param chromecast: Chromecast to cast to. If none is specified, then the default configured Chromecast
            will be used.
        :type chromecast: str

        :param step: Volume decrement between 0 and 100 (default: 100%)
        :type step: float
        """

        chromecast = chromecast or self.chromecast
        cast = self.get_chromecast(chromecast)
        step /= 100
        cast.volume_down(max(step, 0))
        cast.wait()
        return self.status(chromecast=chromecast)

    @action
    def mute(self, chromecast=None):
        """
        Toggle the mute status on the Chromecast

        :param chromecast: Chromecast to cast to. If none is specified, then the default configured Chromecast
            will be used.
        :type chromecast: str
        """

        chromecast = chromecast or self.chromecast
        cast = self.get_chromecast(chromecast)
        cast.set_volume_muted(not cast.status.volume_muted)
        cast.wait()
        return self.status(chromecast=chromecast)


# vim:sw=4:ts=4:et:
